# TODO: ACOPP14 needs human review - complex AC OPF formulation
# from .acopp14 import ACOPP14 as ACOPP14
# TODO: AIRPORT needs human review - constraint values don't match pycutest
# from .airport import AIRPORT as AIRPORT

# TODO: ALLINITA needs human review - L2 group type interpretation issues
# from .allinita import ALLINITA as ALLINITA
# TODO: ALLINITC needs human review - dimension mismatch with pycutest (fixed variables)
# from .allinitc import ALLINITC as ALLINITC
from .alsotame import ALSOTAME as ALSOTAME

# TODO: ANTWERP needs human review - initial value calculation issues
# from .antwerp import ANTWERP as ANTWERP
# TODO: AUG2D needs human review - edge variable structure
# from .aug2d import AUG2D as AUG2D
from .avgasa import AVGASA as AVGASA
from .avgasb import AVGASB as AVGASB

# from .avion2 import AVION2 as AVION2  # TODO: Human review - gradient discrepancies
from .biggsc4 import BIGGSC4 as BIGGSC4
from .bt1 import BT1 as BT1
from .bt2 import BT2 as BT2
from .bt3 import BT3 as BT3
from .bt4 import BT4 as BT4
from .bt5 import BT5 as BT5
from .bt6 import BT6 as BT6
from .bt7 import BT7 as BT7
from .bt8 import BT8 as BT8
from .bt9 import BT9 as BT9
from .bt10 import BT10 as BT10
from .bt11 import BT11 as BT11
from .bt12 import BT12 as BT12
from .bt13 import BT13 as BT13
from .burkehan import BURKEHAN as BURKEHAN
from .byrdsphr import BYRDSPHR as BYRDSPHR
from .cantilvr import CANTILVR as CANTILVR
from .cb2 import CB2 as CB2
from .cb3 import CB3 as CB3
from .chaconn1 import CHACONN1 as CHACONN1
from .chaconn2 import CHACONN2 as CHACONN2

# from .chardis1 import CHARDIS1 as CHARDIS1  # TODO: Human review needed
# from .chardis12 import CHARDIS12 as CHARDIS12  # TODO: Human review needed
# from .cresc4 import CRESC4 as CRESC4  # TODO: Human review - complex crescent area
# TODO: CLNLBEAM needs fixing - dimension mismatch in constraints
# from .clnlbeam import CLNLBEAM as CLNLBEAM
from .cleuven2 import CLEUVEN2 as CLEUVEN2
from .cleuven3 import CLEUVEN3 as CLEUVEN3
from .cleuven4 import CLEUVEN4 as CLEUVEN4
from .cleuven5 import CLEUVEN5 as CLEUVEN5
from .cleuven6 import CLEUVEN6 as CLEUVEN6
from .cleuven7 import CLEUVEN7 as CLEUVEN7

# TODO: CONCON - removed from API due to automatic derivative mismatches with pycutest
# from .concon import CONCON as CONCON
from .coshfun import COSHFUN as COSHFUN
from .csfi1 import CSFI1 as CSFI1
from .csfi2 import CSFI2 as CSFI2
from .dallass import DALLASS as DALLASS
from .deconvc import DECONVC as DECONVC

# TODO: DEGTRIDL needs human review - causes segfault despite correct implementation
# from .degtridl import DEGTRIDL as DEGTRIDL
from .dtoc1l import DTOC1L as DTOC1L
from .dtoc1na import DTOC1NA as DTOC1NA
from .dtoc1nb import DTOC1NB as DTOC1NB
from .dtoc1nc import DTOC1NC as DTOC1NC
from .dtoc1nd import DTOC1ND as DTOC1ND
from .dtoc2 import DTOC2 as DTOC2

# TODO: DTOC3 needs human review - X(1498) and X(1499) treated as 0 by pycutest
# from .dtoc3 import DTOC3 as DTOC3
from .dtoc4 import DTOC4 as DTOC4

# TODO: Human review needed - same constraint issues as EIGENA
# from .eigenaco import EIGENACO as EIGENACO
from .dtoc5 import DTOC5 as DTOC5
from .dtoc6 import DTOC6 as DTOC6
from .elattar import ELATTAR as ELATTAR
from .expfita import EXPFITA as EXPFITA
from .expfitb import EXPFITB as EXPFITB
from .expfitc import EXPFITC as EXPFITC

# TODO: FCCU needs human review - objective value discrepancies
# from .fccu import FCCU as FCCU
# TODO: FEEDLOC needs human review - constraint dimension mismatch
# from .feedloc import FEEDLOC as FEEDLOC
from .fletcher import FLETCHER as FLETCHER
from .flt import FLT as FLT
from .gigomez2 import GIGOMEZ2 as GIGOMEZ2
from .hadamard import HADAMARD as HADAMARD

# TODO: Human review needed - parameter/dimension mismatch with pycutest
# JAX implementation uses correct SIF parameters (N=10 → 21 variables)
# But pycutest expects 5001 variables, suggesting different problem variant
# Issue persists even with drop_fixed_variables=False
from .hager1 import HAGER1 as HAGER1
from .hager2 import HAGER2 as HAGER2

# TODO: HAGER3 needs human review - marked for future import
# from .hager3 import HAGER3 as HAGER3
from .hager4 import HAGER4 as HAGER4

# TODO: Human review needed - complex SIF structure
from .haifal import HAIFAL as HAIFAL
from .haifam import (
    HAIFAM as HAIFAM,
)
from .haifas import HAIFAS as HAIFAS
from .himmelbc import HIMMELBC as HIMMELBC
from .himmelbd import HIMMELBD as HIMMELBD
from .himmelbe import HIMMELBE as HIMMELBE

# from .himmelp2 import HIMMELP2 as HIMMELP2  # TODO: Human review - OBNL issues
# from .himmelp3 import HIMMELP3 as HIMMELP3  # TODO: Human review - OBNL issues
# from .himmelp4 import HIMMELP4 as HIMMELP4  # TODO: Human review - OBNL issues
# from .himmelp5 import HIMMELP5 as HIMMELP5  # TODO: Human review - OBNL issues
# from .himmelp6 import HIMMELP6 as HIMMELP6  # TODO: Human review - OBNL issues
from .hs6 import HS6 as HS6
from .hs7 import HS7 as HS7
from .hs8 import HS8 as HS8
from .hs9 import HS9 as HS9
from .hs10 import HS10 as HS10
from .hs11 import HS11 as HS11
from .hs12 import HS12 as HS12
from .hs13 import HS13 as HS13
from .hs14 import HS14 as HS14
from .hs15 import HS15 as HS15
from .hs16 import HS16 as HS16
from .hs17 import HS17 as HS17
from .hs18 import HS18 as HS18
from .hs19 import HS19 as HS19
from .hs20 import HS20 as HS20
from .hs21 import HS21 as HS21
from .hs22 import HS22 as HS22
from .hs23 import HS23 as HS23
from .hs24 import HS24 as HS24
from .hs26 import HS26 as HS26
from .hs27 import HS27 as HS27
from .hs28 import HS28 as HS28
from .hs29 import HS29 as HS29
from .hs30 import HS30 as HS30
from .hs31 import HS31 as HS31
from .hs32 import HS32 as HS32
from .hs33 import HS33 as HS33
from .hs34 import HS34 as HS34
from .hs35 import HS35 as HS35
from .hs36 import HS36 as HS36
from .hs37 import HS37 as HS37
from .hs39 import HS39 as HS39
from .hs40 import HS40 as HS40
from .hs41 import HS41 as HS41
from .hs42 import HS42 as HS42
from .hs43 import HS43 as HS43
from .hs44 import HS44 as HS44
from .hs46 import HS46 as HS46
from .hs47 import HS47 as HS47
from .hs48 import HS48 as HS48
from .hs49 import HS49 as HS49
from .hs50 import HS50 as HS50
from .hs51 import HS51 as HS51
from .hs52 import HS52 as HS52
from .hs53 import HS53 as HS53
from .hs54 import HS54 as HS54
from .hs55 import HS55 as HS55
from .hs56 import HS56 as HS56
from .hs57 import HS57 as HS57

# from .hs59 import HS59 as HS59  # TODO: Human review - objective function discrepancy
from .hs60 import HS60 as HS60
from .hs61 import HS61 as HS61
from .hs62 import HS62 as HS62
from .hs63 import HS63 as HS63
from .hs64 import HS64 as HS64
from .hs65 import HS65 as HS65
from .hs66 import HS66 as HS66

# from .hs67 import HS67 as HS67  # TODO: Human review - several discrepancies
from .hs68 import HS68 as HS68
from .hs69 import HS69 as HS69

# TODO: HS70 needs human review - test failures
# from .hs70 import HS70 as HS70
from .hs71 import HS71 as HS71
from .hs72 import HS72 as HS72
from .hs73 import HS73 as HS73

# TODO: HS74 needs human review - constraint Jacobian values differ by large factors
# from .hs74 import HS74 as HS74
# TODO: HS75 needs human review - same issues as HS74
# from .hs75 import HS75 as HS75
from .hs76 import HS76 as HS76
from .hs77 import HS77 as HS77
from .hs78 import HS78 as HS78
from .hs79 import HS79 as HS79
from .hs80 import HS80 as HS80
from .hs81 import HS81 as HS81
from .hs83 import HS83 as HS83

# TODO: HS84 needs human review - objective value discrepancy (~2%)
# from .hs84 import HS84 as HS84
from .hs93 import HS93 as HS93

# from .hs99 import HS99 as HS99  # TODO: Needs human review - complex recursion
from .hs100 import HS100 as HS100
from .hs101 import HS101 as HS101
from .hs102 import HS102 as HS102
from .hs103 import HS103 as HS103
from .hs104 import HS104 as HS104
from .hs105 import HS105 as HS105
from .hs106 import HS106 as HS106
from .hs107 import HS107 as HS107
from .hs108 import HS108 as HS108

# from .hs109 import HS109 as HS109  # TODO: Human review - sign convention issues
from .hs111 import HS111 as HS111
from .hs112 import HS112 as HS112
from .hs113 import HS113 as HS113
from .hs114 import HS114 as HS114
from .hs116 import HS116 as HS116
from .hs117 import HS117 as HS117

# TODO: HS118 needs human review - constraint Jacobian ordering mismatch
# from .hs118 import HS118 as HS118
from .hs119 import HS119 as HS119
from .hydroell import HYDROELL as HYDROELL

# TODO: KISSING needs human review - runtime issue (5.37x, slightly over 5x threshold)
# from .kissing import KISSING as KISSING
# from .kissing2 import KISSING2 as KISSING2  # TODO: Human review needed
# TODO: Human review needed - KIWCRESC constraint values differ by 2.0 from pycutest
# from .kiwcresc import KIWCRESC as KIWCRESC
# TODO: Human review needed - KSIP needs vectorization, dtype promotion errors with JAX
# from .ksip import KSIP as KSIP
from .lootsma import LOOTSMA as LOOTSMA
from .lukvle1 import LUKVLE1 as LUKVLE1

# from .lukvle2 import LUKVLE2 as LUKVLE2
from .lukvle3 import LUKVLE3 as LUKVLE3

# from .lukvle4 import LUKVLE4 as LUKVLE4  # Use LUKVLE4C instead
# from .lukvle4c import LUKVLE4C as LUKVLE4C
from .lukvle5 import LUKVLE5 as LUKVLE5
from .lukvle6 import LUKVLE6 as LUKVLE6
from .lukvle7 import LUKVLE7 as LUKVLE7
from .lukvle8 import LUKVLE8 as LUKVLE8

# from .lukvle9 import LUKVLE9 as LUKVLE9  # TODO: Human review needed - Jacobian issues
from .lukvle10 import LUKVLE10 as LUKVLE10
from .lukvle11 import LUKVLE11 as LUKVLE11

# from .lukvle12 import LUKVLE12 as LUKVLE12  # Has constraint function inconsistencies
from .lukvle13 import LUKVLE13 as LUKVLE13
from .lukvle14 import LUKVLE14 as LUKVLE14
from .lukvle15 import LUKVLE15 as LUKVLE15
from .lukvle16 import LUKVLE16 as LUKVLE16
from .lukvle17 import LUKVLE17 as LUKVLE17
from .lukvle18 import LUKVLE18 as LUKVLE18
from .lukvli1 import LUKVLI1 as LUKVLI1

# from .lukvli2 import LUKVLI2 as LUKVLI2
from .lukvli3 import LUKVLI3 as LUKVLI3

# from .lukvli4 import LUKVLI4 as LUKVLI4  # Use LUKVLI4C instead
# from .lukvli4c import LUKVLI4C as LUKVLI4C
from .lukvli5 import LUKVLI5 as LUKVLI5
from .lukvli6 import LUKVLI6 as LUKVLI6
from .lukvli7 import LUKVLI7 as LUKVLI7
from .lukvli8 import LUKVLI8 as LUKVLI8

# from .lukvli9 import LUKVLI9 as LUKVLI9  # TODO: Human review needed - Jacobian issues
from .lukvli10 import LUKVLI10 as LUKVLI10
from .lukvli11 import LUKVLI11 as LUKVLI11

# from .lukvli12 import LUKVLI12 as LUKVLI12  # Has constraint function inconsistencies
from .lukvli13 import LUKVLI13 as LUKVLI13
from .lukvli14 import LUKVLI14 as LUKVLI14
from .lukvli15 import LUKVLI15 as LUKVLI15
from .lukvli16 import LUKVLI16 as LUKVLI16
from .lukvli17 import LUKVLI17 as LUKVLI17
from .lukvli18 import LUKVLI18 as LUKVLI18
from .makela1 import MAKELA1 as MAKELA1
from .makela2 import MAKELA2 as MAKELA2
from .makela3 import MAKELA3 as MAKELA3
from .makela4 import MAKELA4 as MAKELA4
from .maratos import MARATOS as MARATOS
from .mss1 import MSS1 as MSS1
from .mss2 import MSS2 as MSS2
from .mss3 import MSS3 as MSS3
from .odfits import ODFITS as ODFITS
from .oet1 import OET1 as OET1
from .oet2 import OET2 as OET2
from .oet3 import OET3 as OET3
from .oet4 import OET4 as OET4
from .oet5 import OET5 as OET5
from .oet6 import OET6 as OET6
from .oet7 import OET7 as OET7
from .optcdeg2 import OPTCDEG2 as OPTCDEG2
from .optcdeg3 import OPTCDEG3 as OPTCDEG3
from .optcntrl import OPTCNTRL as OPTCNTRL
from .optctrl3 import OPTCTRL3 as OPTCTRL3
from .optctrl6 import OPTCTRL6 as OPTCTRL6
from .optmass import OPTMASS as OPTMASS
from .optprloc import OPTPRLOC as OPTPRLOC

# from .orthrdm2 import ORTHRDM2 as ORTHRDM2  # TODO: Human review - singular Jacobian
# from .orthrds2 import ORTHRDS2 as ORTHRDS2  # TODO: Human review - singular Jacobian
from .orthrds2c import ORTHRDS2C as ORTHRDS2C

# from .orthrega import ORTHREGA as ORTHREGA  # TODO: Human review - formulation diffs
from .orthregb import ORTHREGB as ORTHREGB
from .orthregc import ORTHREGC as ORTHREGC
from .orthregd import ORTHREGD as ORTHREGD
from .orthrege import ORTHREGE as ORTHREGE
from .orthregf import ORTHREGF as ORTHREGF
from .orthrgdm import ORTHRGDM as ORTHRGDM
from .orthrgds import ORTHRGDS as ORTHRGDS
from .pentagon import PENTAGON as PENTAGON
from .polak1 import POLAK1 as POLAK1
from .polak2 import POLAK2 as POLAK2
from .polak3 import POLAK3 as POLAK3
from .polak4 import POLAK4 as POLAK4
from .polak5 import POLAK5 as POLAK5
from .polak6 import POLAK6 as POLAK6

# from .polygon import POLYGON as POLYGON  # TODO: Human review - sign conventions
from .reading1 import READING1 as READING1
from .reading2 import READING2 as READING2
from .reading3 import READING3 as READING3
from .reading4 import READING4 as READING4
from .reading5 import READING5 as READING5

# from .reading6 import READING6 as READING6  # TODO: Human review needed
# Note: READING7 and READING8 exist but are not implemented due to a CUTEst bug:
# the starting point is the solution too
from .reading9 import READING9 as READING9
from .simpllpa import SIMPLLPA as SIMPLLPA
from .simpllpb import SIMPLLPB as SIMPLLPB

# from .sinrosnb import SINROSNB as SINROSNB  # TODO: Human review - scaling issues
from .sipow1 import SIPOW1 as SIPOW1
from .sipow2 import SIPOW2 as SIPOW2

# TODO: Human review needed - constraint test failures (vectorized but sign issues)
# from .spin2op import SPIN2OP as SPIN2OP
# TODO: Human review needed - constraint issues with auxiliary variables
# from .spinop import SPINOP as SPINOP
# TODO: STEENBRB needs human review - gradient test failing
# from .steenbrb import STEENBRB as STEENBRB
# TODO: SIPOW3 needs human review - constraint formulation issues
# from .sipow3 import SIPOW3 as SIPOW3
# TODO: SIPOW4 needs human review - constraint formulation issues
# from .sipow4 import SIPOW4 as SIPOW4
# TODO: TAX13322 needs human review - complex objective structure, off by ~80x
# from .tax13322 import TAX13322 as TAX13322
from .tenbars1 import TENBARS1 as TENBARS1
from .tenbars2 import TENBARS2 as TENBARS2
from .tenbars3 import TENBARS3 as TENBARS3
from .tenbars4 import TENBARS4 as TENBARS4
from .tro3x3 import TRO3X3 as TRO3X3
from .tro4x4 import TRO4X4 as TRO4X4
from .tro5x5 import TRO5X5 as TRO5X5
from .tro6x2 import TRO6X2 as TRO6X2
from .tro11x3 import TRO11X3 as TRO11X3
from .tro21x5 import TRO21X5 as TRO21X5
from .tro41x9 import TRO41X9 as TRO41X9

# TODO: TRUSPYR1 needs human review - complex constraint scaling issues
# from .truspyr1 import TRUSPYR1 as TRUSPYR1
# TODO: TRUSPYR2 needs human review - test requested to be removed
# from .truspyr2 import TRUSPYR2 as TRUSPYR2
# from .vanderm3 import VANDERM3 as VANDERM3  # TODO: Human review needed
# from .vanderm4 import VANDERM4 as VANDERM4  # TODO: Human review needed
# TODO: TWIR problems need human review - complex trilinear constraint formulation
# from .twirism1 import TWIRISM1 as TWIRISM1
# from .twirimd1 import TWIRIMD1 as TWIRIMD1
# from .twiribg1 import TWIRIBG1 as TWIRIBG1
from .zecevic2 import ZECEVIC2 as ZECEVIC2
from .zecevic3 import ZECEVIC3 as ZECEVIC3
from .zecevic4 import ZECEVIC4 as ZECEVIC4


constrained_minimisation_problems = (
    # ACOPP14(),  # TODO: needs human review - complex AC OPF formulation
    # AIRPORT(),  # TODO: Human review - constraint values don't match pycutest
    # ALLINITA(),  # TODO: needs human review - L2 group type interpretation
    # ALLINITC(),  # Human review needed - dimension mismatch
    ALSOTAME(),
    TRO3X3(),
    TRO4X4(),
    TRO5X5(),
    TRO6X2(),
    TRO11X3(),
    TRO21X5(),
    TRO41X9(),
    # ANTWERP(),  # TODO: needs human review - initial value calculation
    AVGASA(),
    AVGASB(),
    # DEGTRIDL(),  # TODO: Human review - causes segfault despite correct implementation
    # AVION2(),  # TODO: Human review - gradient discrepancies
    BIGGSC4(),
    BURKEHAN(),
    BYRDSPHR(),
    CANTILVR(),
    CB2(),
    CB3(),
    CHACONN1(),
    CHACONN2(),
    # CHARDIS1(),  # TODO: Human review needed
    # CHARDIS12(),  # TODO: Human review needed
    # CLNLBEAM(),  # TODO: Dimension mismatch in constraints
    CLEUVEN2(),
    CLEUVEN3(),
    CLEUVEN4(),
    CLEUVEN5(),
    CLEUVEN6(),
    CLEUVEN7(),
    # CONCON(),  # TODO: Removed - automatic derivative mismatches
    COSHFUN(),
    # CRESC4(),  # TODO: Human review - complex crescent area formula
    CSFI1(),
    CSFI2(),
    DALLASS(),
    DECONVC(),
    DTOC1L(),
    DTOC1NA(),
    DTOC1NB(),
    DTOC1NC(),
    DTOC1ND(),
    DTOC2(),
    # DTOC3(),  # Human review needed
    DTOC4(),
    DTOC5(),
    DTOC6(),
    # EIGENACO(),  # TODO: Human review needed - same constraint issues as EIGENA
    ELATTAR(),
    # EXPFITA(),  # TODO: Human review - fundamental formulation differences
    # EXPFITB(),  # TODO: Human review - fundamental formulation differences
    # EXPFITC(),  # TODO: Human review - fundamental formulation differences
    # FCCU(),  # TODO: FCCU needs human review - objective value discrepancies
    # FEEDLOC(),  # TODO: FEEDLOC needs human review - constraint dimension mismatch
    FLETCHER(),
    FLT(),
    GIGOMEZ2(),
    HADAMARD(),
    HAGER1(),
    HAGER2(),
    # HAGER3(),  # TODO: Human review needed - marked for future import
    HAGER4(),
    HAIFAS(),
    HAIFAM(),  # TODO: Human review needed - complex SIF structure
    HAIFAL(),
    HS6(),
    HS7(),
    HS8(),
    HS9(),
    HS10(),
    HS11(),
    HS12(),
    HS13(),
    HS14(),
    HS15(),
    HS16(),
    HS17(),
    HS18(),
    HS19(),
    HS20(),
    HS21(),
    HS22(),
    HS23(),
    HS24(),
    HS26(),
    HS27(),
    HS28(),
    HS29(),
    HS30(),
    HS31(),
    HS32(),
    HS33(),
    HS34(),
    HS35(),
    HS36(),
    HS37(),
    HS39(),
    HS40(),
    HS41(),
    HS42(),
    HS43(),
    HS44(),
    HS46(),
    HS47(),
    HS48(),
    HS49(),
    HS50(),
    HS51(),
    HS52(),
    HS53(),
    HS54(),
    HS55(),
    HS56(),
    HS57(),
    # HS59(),  # TODO: Human review - objective function discrepancy
    HS60(),
    HS61(),
    HS62(),
    HS63(),
    HS64(),
    HS65(),
    HS66(),
    # HS67(),  # TODO: Human review - several discrepancies
    HS68(),
    HS69(),
    HS71(),
    HS72(),
    HS73(),
    # HS74(),  # Human review needed - constraint Jacobian issues
    # HS75(),  # Human review needed - same issues as HS74
    HS76(),
    HS77(),
    HS78(),
    HS79(),
    HS80(),
    HS81(),
    HS83(),
    HS93(),
    # HS99(),  # TODO: Needs human review - complex recursive formulation
    HS100(),
    HS101(),
    HS102(),
    HS103(),
    HS104(),
    HS105(),
    HS106(),
    HS107(),
    HS108(),
    # HS109(),  # TODO: Human review needed - sign convention issues
    HS111(),
    HS112(),
    HS113(),
    HS114(),
    HS116(),
    HS117(),
    # HS118(),  # TODO: Human review - constraint Jacobian ordering mismatch
    HS119(),
    HYDROELL(),
    # KIWCRESC(),  # TODO: Human review - constraint values differ by 2.0 from pycutest
    # KISSING(),  # TODO: Human review - runtime issue (5.37x)
    # KISSING2(),  # TODO: Human review needed
    # KSIP(),  # TODO: Human review needed - needs vectorization, dtype promotion errors
    LOOTSMA(),
    HIMMELBC(),
    HIMMELBD(),
    HIMMELBE(),
    # HIMMELP2(),  # TODO: Human review needed - OBNL element issues
    # HIMMELP3(),  # TODO: Human review needed - OBNL element issues
    # HIMMELP4(),  # TODO: Human review needed - OBNL element issues
    # HIMMELP5(),  # TODO: Human review needed - OBNL element issues
    # HIMMELP6(),  # TODO: Human review needed - OBNL element issues
    HYDROELL(),
    LOOTSMA(),
    MARATOS(),
    MSS1(),
    MSS2(),
    MSS3(),
    ODFITS(),
    OET1(),
    OET2(),
    OET3(),
    OET4(),
    OET5(),
    OET6(),
    OET7(),
    OPTCDEG2(),
    OPTCDEG3(),
    OPTCNTRL(),
    OPTCTRL3(),
    OPTCTRL6(),
    OPTMASS(),
    OPTPRLOC(),
    # ORTHRDM2(),  # TODO: Human review - singular Jacobian issues
    # ORTHRDS2(),  # TODO: Human review - singular Jacobian issues
    ORTHRDS2C(),
    # ORTHREGA(),  # TODO: Human review - fundamental formulation differences
    ORTHREGB(),
    ORTHREGC(),
    ORTHREGD(),
    ORTHREGE(),
    ORTHREGF(),
    ORTHRGDM(),
    ORTHRGDS(),
    PENTAGON(),
    POLAK1(),
    POLAK2(),
    POLAK3(),
    POLAK4(),
    POLAK5(),
    POLAK6(),
    # POLYGON(),  # TODO: Human review - constraint sign convention differences
    READING1(),
    READING2(),
    READING3(),
    READING4(),
    READING5(),
    # READING6(),  # TODO: Human review needed
    # Note: READING7 and READING8 exist but are not implemented due to a CUTEst bug:
    # the starting point is the solution too
    READING9(),
    SIMPLLPA(),
    SIMPLLPB(),
    # SINROSNB(),  # TODO: Human review - objective scaling issues
    SIPOW1(),
    SIPOW2(),
    # TAX13322(),  # TODO: Human review - complex objective structure
    TENBARS1(),
    TENBARS2(),
    TENBARS3(),
    TENBARS4(),
    # SPINOP(),  # TODO: Human review - constraint issues with auxiliary variables
    # SPIN2OP(),  # TODO: Human review - constraint test failures
    # STEENBRB(),  # TODO: Human review - gradient test failing
    # SIPOW3(),  # TODO: Human review - constraint formulation issues
    # SIPOW4(),  # TODO: Human review - constraint formulation issues
    # VANDERM1(),  # Moved to nonlinear equations (NOR2 classification)
    # VANDERM2(),  # Moved to nonlinear equations (NOR2 classification)
    # VANDERM3(),  # TODO: Human review - constraint values don't match
    # VANDERM4(),  # TODO: Human review - constraint values don't match
    MAKELA1(),
    MAKELA2(),
    MAKELA3(),
    MAKELA4(),
    # HS70(),  # TODO: Human review - test failures
    # HS84(),  # TODO: Human review - objective value discrepancy
    # TODO: TWIR problems need human review - complex trilinear constraint formulation
    # TWIRISM1(),
    # TWIRIMD1(),
    # TWIRIBG1(),
    ZECEVIC2(),
    ZECEVIC3(),
    ZECEVIC4(),
    # TRUSPYR1(),  # TODO: Human review - complex constraint scaling issues
    # TRUSPYR2(),  # TODO: Human review - test requested to be removed
    BT1(),
    BT2(),
    BT3(),
    BT4(),
    BT5(),
    BT6(),
    BT7(),
    BT8(),
    BT9(),
    BT10(),
    BT11(),
    BT12(),
    BT13(),
    LUKVLE1(),
    # LUKVLE2(),
    LUKVLE3(),
    # LUKVLE4(),  # Has factor ~2380 error due to SIF bug, use LUKVLE4C instead
    # LUKVLE4C(),  # Has 3% discrepancy with pycutest
    LUKVLE5(),
    LUKVLE6(),
    LUKVLE7(),
    LUKVLE8(),
    # LUKVLE9(),  # TODO: Human review needed - Jacobian issues
    LUKVLE10(),
    LUKVLE11(),
    # LUKVLE12(),  # Has constraint function inconsistencies
    LUKVLE13(),
    LUKVLE14(),
    LUKVLE15(),
    LUKVLE16(),
    LUKVLE17(),
    LUKVLE18(),
    LUKVLI1(),
    # LUKVLI2(),
    LUKVLI3(),
    # LUKVLI4(),  # Has factor ~2380 error due to SIF bug, use LUKVLI4C instead
    # LUKVLI4C(),  # Has 3% discrepancy with pycutest
    LUKVLI5(),
    LUKVLI6(),
    LUKVLI7(),
    LUKVLI8(),
    # LUKVLI9(),  # TODO: Human review needed - Jacobian issues
    LUKVLI10(),
    LUKVLI11(),
    # LUKVLI12(),  # Has constraint function inconsistencies
    LUKVLI13(),
    LUKVLI14(),
    LUKVLI15(),
    LUKVLI16(),
    LUKVLI17(),
    LUKVLI18(),
)



"""
A nonlinear minmax problem in eleven variables.

Note: the original statement of the problem contains an inconsistent
index i. This has been replaced by 1, assuming a very common typo.
But the optimal solution of the resulting problem differs from that
quoted in the source.

Source:
E. Polak, D.H. Mayne and J.E. Higgins,
"Superlinearly convergent algorithm for min-max problems"
JOTA 69, pp. 407-439, 1991.

SIF input: Ph. Toint, Nov 1993.

classification LOR2-AN-12-10
"""

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class POLAK3(AbstractConstrainedMinimisation):
    @property
    def name(self) -> str:
        return "POLAK3"

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 12  # 11 X variables + 1 U variable
    m: int = 10  # 10 inequality constraints

    @property
    def y0(self):
        # All variables start at 1.0
        return jnp.ones(self.n)

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args
        # Objective is simply U (the 12th variable)
        u = y[11]
        return jnp.array(u)

    def constraint(self, y):
        # Variables: X(1) to X(11) are y[0] to y[10], U is y[11]
        x = y[:11]
        u = y[11]

        # F(I) constraints: U - sum_j (1/j) * E(I,J) >= 0
        # where E(I,J) = exp((X(J) - sin((I-1) + 2*J))^2)
        # But pycutest expects -F(I) <= 0, so we return the negative

        inequality_constraints = []

        for i in range(1, 11):  # I from 1 to 10
            f_i = u
            for j in range(1, 12):  # J from 1 to 11
                # Element E(I,J) with A = I-1, B = J-1
                # From SIF lines 68-69: ZP E(I,J) A RI-1, ZP E(I,J) B RJ-1
                # But in lines 67-68, RJ-1 = J (not J-1)
                a = float(i - 1)
                b = float(j)  # B = J, not J-1
                v = x[j - 1] - jnp.sin(a + b + b)  # XX - sin(A + B + B)
                e_ij = jnp.exp(v * v)
                f_i = f_i - (1.0 / j) * e_ij
            # Pycutest expects -F(I) <= 0 instead of F(I) >= 0
            inequality_constraints.append(-f_i)

        return None, jnp.array(inequality_constraints)

    @property
    def bounds(self):
        # All variables are free (unbounded)
        return None

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # According to the SIF file comment (line 90),
        # the optimal objective value is 5.93300252
        return jnp.array(5.93300252)



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class PALMER5B(AbstractBoundedMinimisation):
    """A nonlinear least squares problem with bounds arising from chemical kinetics.

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 + A2 X**2 + A4 X**4 + A6 X**6 + A8 X**8 + A10 X**10
                + A12 X**12 + B / ( C + X**2 ), B, C nonnegative.

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1992.

    classification SBR2-RN-9-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Data points
    M: int = 23

    # X values (radians) - starting from index 12
    X_data = jnp.array(
        [
            0.000000,
            1.570796,
            1.396263,
            1.308997,
            1.221730,
            1.125835,
            1.047198,
            0.872665,
            0.698132,
            0.523599,
            0.349066,
            0.174533,
        ]
    )

    # Y values (KJmol-1)
    Y_data = jnp.array(
        [
            83.57418,
            81.007654,
            18.983286,
            8.051067,
            2.044762,
            0.000000,
            1.170451,
            10.479881,
            25.785001,
            44.126844,
            62.822177,
            77.719674,
        ]
    )

    @property
    def n(self):
        """Number of variables."""
        return 9  # A0, A2, A4, A6, A8, A10, A12, B, C

    def objective(self, y, args):
        """Compute the objective function (least squares)."""
        del args

        # Extract variables
        A0, A2, A4, A6, A8, A10, A12, B, C = y

        # Precompute powers of X
        X_sqr = self.X_data * self.X_data
        X_4 = X_sqr * X_sqr
        X_6 = X_sqr * X_4
        X_8 = X_sqr * X_6
        X_10 = X_sqr * X_8
        X_12 = X_sqr * X_10

        # Model predictions
        predictions = (
            A0
            + A2 * X_sqr
            + A4 * X_4
            + A6 * X_6
            + A8 * X_8
            + A10 * X_10
            + A12 * X_12
            + B / (C + X_sqr)
        )

        # Residuals
        residuals = predictions - self.Y_data

        # Sum of squares
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        """Initial guess."""
        return jnp.ones(9)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return jnp.zeros(9)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(4.0606141e-02)

    @property
    def bounds(self):
        """Returns the bounds on the variable y."""
        # A0-A12 are free, B and C have lower bounds
        lower = jnp.array(
            [
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                0.00001,
                0.00001,
            ]
        )
        upper = jnp.full(9, jnp.inf)
        return lower, upper



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


# TODO: Human review needed
# Attempts made: Vectorized the implementation
# Suspected issues: SIF file bug on line 64 - "X(I)" should be "X(N)"
# Resources needed: Fix in the SIF file or confirm intended behavior
class MOREBVNE(AbstractNonlinearEquations):
    """
    The Boundary Value problem.
    This is the nonlinear least-squares version without fixed variables.

    Source:  problem 28 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#17 (p. 75).

    SIF input: Ph. Toint, Dec 1989 and Nick Gould, Oct 1992.
    Modification as a set of nonlinear equations: Nick Gould, Oct 2015.

    classification NOR2-MN-V-V
    """

    n: int = 10
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, n: int = 10):
        self.n = n

    def num_residuals(self) -> int:
        """Number of residuals equals number of variables."""
        return self.n

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        n = self.n
        h = 1.0 / (n + 1)

        # x[i] = t[i] * (t[i] - 1) where t[i] = i*h
        i_indices = jnp.arange(1, n + 1, dtype=jnp.float64)
        t = i_indices * h
        x0 = t * (t - 1.0)

        return x0

    def _wcube_element(self, v: Array, b: float) -> Array:
        """WCUBE element: (v + b)^3."""
        vplusb = v + b
        return vplusb**3

    def residual(self, y: Array, args) -> Array:
        """Compute the residual vector."""
        n = self.n
        h = 1.0 / (n + 1)
        h2 = h * h
        halfh2 = 0.5 * h2

        # Create arrays for vectorized computation
        i_indices = jnp.arange(1, n + 1, dtype=jnp.float64)
        ih_plus_1 = i_indices * h + 1.0

        # Compute all E(i) values: (X(i) + (i*h + 1))^3
        e_values = (y + ih_plus_1) ** 3

        # Initialize residuals
        residuals = jnp.zeros(n, dtype=y.dtype)

        # G(1): 2*X(1) - X(2) + halfh2 * E(1)
        residuals = residuals.at[0].set(2.0 * y[0] - y[1] + halfh2 * e_values[0])

        # G(i) for i = 2 to N-1: -X(i-1) + 2*X(i) - X(i+1) + halfh2 * E(i)
        residuals = residuals.at[1 : n - 1].set(
            -y[0 : n - 2] + 2.0 * y[1 : n - 1] - y[2:n] + halfh2 * e_values[1 : n - 1]
        )

        # G(N): -X(N-1) + 2*X(N) + halfh2 * E(N)
        residuals = residuals.at[n - 1].set(
            -y[n - 2] + 2.0 * y[n - 1] + halfh2 * e_values[n - 1]
        )

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Not explicitly given, but for nonlinear equations should satisfy F(x*) = 0
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """Bounds for variables - free variables."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed
# Attempts made: Multiple attempts to handle fixed variables and constraint signs
# Suspected issues: pycutest vs sif2jax convention differences for fixed variables
#                   and constraint signs
# Resources needed: Understanding of how pycutest handles fixed boundary values
class POLYGON(AbstractConstrainedMinimisation):
    """Find the polygon of maximal area with fixed diameter.

    Find the polygon of maximal area, among polygons with nv sides and
    diameter d <= 1.

    This is problem 1 in the COPS (Version 2) collection of
    E. Dolan and J. More'
    see "Benchmarking Optimization Software with COPS"
    Argonne National Labs Technical Report ANL/MCS-246 (2000)

    SIF input: Nick Gould, December 2000

    Classification: OOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Parameters
    NV: int = 100  # Number of vertices

    @property
    def n(self):
        """Number of variables: 2*NV (including fixed R(NV) and THETA(NV))."""
        return 2 * self.NV

    @property
    def y0(self):
        """Initial guess."""
        nv = self.NV
        nv_plus_1 = nv + 1.0
        nv_plus_1_sq = nv_plus_1 * nv_plus_1
        ratr = nv_plus_1_sq / 4.0
        pi = jnp.pi
        ratt = pi / nv

        y0 = jnp.zeros(self.n)

        # Set initial R and THETA values (interleaved: r1, theta1, r2, theta2, ...)
        # including fixed R(NV) and THETA(NV)
        for i in range(
            nv - 1
        ):  # 0 to NV-2 in 0-based, corresponds to 1 to NV-1 in 1-based
            i_1based = i + 1
            ratri = nv_plus_1 - i_1based  # RATRI = NV+1 - I
            ratri = ratri * i_1based  # RATRI = RATRI * I
            ratri = ratri / ratr  # RATRI = RATRI / RATR
            ratti = ratt * i_1based  # RATTI = RATT * I

            y0 = y0.at[2 * i].set(ratri)  # R(i+1)
            y0 = y0.at[2 * i + 1].set(ratti)  # THETA(i+1)

        # Set fixed values for R(NV) and THETA(NV)
        y0 = y0.at[2 * (nv - 1)].set(0.0)  # R(NV) = 0.0
        y0 = y0.at[2 * (nv - 1) + 1].set(pi)  # THETA(NV) = PI

        return y0

    @property
    def args(self):
        """No additional arguments."""
        return None

    def objective(self, y, args):
        """Compute the objective function (negative polygon area).

        Area = -0.5 * sum_{i=1}^{NV-1} r[i+1]*r[i]*sin(theta[i+1] - theta[i])
        """
        del args  # Not used

        nv = self.NV

        # Extract interleaved variables - vectorized
        indices = jnp.arange(nv)
        r = y[2 * indices]  # R values (R(1) to R(NV))
        theta = y[2 * indices + 1]  # THETA values (THETA(1) to THETA(NV))

        # Vectorized computation over i from 0 to NV-2
        r1 = r[:-1]  # r[i] for i = 0 to NV-2
        r2 = r[1:]  # r[i+1] for i = 0 to NV-2
        t1 = theta[1:]  # theta[i+1] for i = 0 to NV-2
        t2 = theta[:-1]  # theta[i] for i = 0 to NV-2

        # SI element: r1 * r2 * sin(t1 - t2)
        area_terms = -0.5 * r1 * r2 * jnp.sin(t1 - t2)
        area = jnp.sum(area_terms)

        return area

    def constraint(self, y):
        """Compute the constraints.

        Inequality constraints:
        - Order constraints: theta[i+1] >= theta[i] for i = 1 to NV-1
        - Distance constraints:
          r[i]^2 + r[j]^2 - 2*r[i]*r[j]*cos(theta[j] - theta[i]) <= 1
          for all i < j
        """
        nv = self.NV

        # Extract interleaved variables - vectorized
        indices = jnp.arange(nv)
        r = y[2 * indices]  # R values (R(1) to R(NV))
        theta = y[2 * indices + 1]  # THETA values (THETA(1) to THETA(NV))

        # Order constraints (NV-1 constraints) - vectorized
        order_constraints = theta[1:] - theta[:-1]  # theta[i+1] - theta[i] >= 0

        # Distance constraints - compute all pairs (i, j) where i < j
        # Use jnp.triu_indices to get upper triangular indices
        i_vals, j_vals = jnp.triu_indices(nv, k=1)

        # Compute distance constraints for all valid pairs
        r_i = r[i_vals]
        r_j = r[j_vals]
        theta_i = theta[i_vals]
        theta_j = theta[j_vals]

        dist_sq = r_i**2 + r_j**2 - 2.0 * r_i * r_j * jnp.cos(theta_j - theta_i)
        distance_constraints = dist_sq - 1.0  # <= 1 constraint

        # No equality constraints
        equalities = None

        # All constraints are inequalities (>= 0)
        inequalities = jnp.concatenate([order_constraints, distance_constraints])

        return equalities, inequalities

    @property
    def bounds(self):
        """Bounds on variables."""
        nv = self.NV
        lower = jnp.zeros(self.n)
        upper = jnp.ones(self.n)

        # Variables are interleaved: r1, theta1, r2, theta2, ...
        # R(i) in [0, 1] and THETA(i) in [0, PI] for i = 1 to NV
        indices = jnp.arange(nv)
        upper = upper.at[2 * indices].set(1.0)  # R(i)
        upper = upper.at[2 * indices + 1].set(jnp.pi)  # THETA(i)

        # Fix R(NV) = 0.0 and THETA(NV) = PI
        lower = lower.at[2 * (nv - 1)].set(0.0)
        upper = upper.at[2 * (nv - 1)].set(0.0)
        lower = lower.at[2 * (nv - 1) + 1].set(jnp.pi)
        upper = upper.at[2 * (nv - 1) + 1].set(jnp.pi)

        return lower, upper

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value for NV=100."""
        return jnp.array(-0.77847)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Human review needed
# Attempts made:
# 1. Added missing factor of 0.5 to objective function
# 2. Fixed E(k) equations from 10*(X^2 - 10*X) to 10*X^2 - 10*X
# 3. Verified equation indexing and interleaving
# 4. Confirmed objective value matches expected (2.0 at zeros for N=5)
# Suspected issues:
# - Gradient computation shows sparse pattern [1,0,0,0,0] at zeros
# - Numerical vs analytical gradient mismatch (~0.596 vs 1.0)
# - Possible precision or formulation differences with pycutest
# Resources needed:
# - Direct comparison with Fortran source code
# - Verification of gradient computation methodology in pycutest
# - Check if L2 group type handling differs between implementations
class LUKSAN22LS(AbstractUnconstrainedMinimisation):
    """Problem 22 (attracting-repelling) in the paper L. Luksan: Hybrid methods in
    large sparse nonlinear least squares. J. Optimization Theory and Applications 89,
    pp. 575-595, 1996.

    This is a large sparse nonlinear least squares problem with exponential terms,
    designed to test optimization algorithms on problems with attracting-repelling
    behavior due to the combination of exponential terms with different signs.

    Note: The implementation follows pycutest conventions which differ from the
    original SIF file in three ways:
    1. E(1) uses X(1) - 1.0 instead of X(1) + 1.0
    2. E(k+1) uses addition between exponential terms instead of subtraction
    3. E(2N-2) omits the -10.0*X(N) term

    Source: Luksan, L. (1996)
    Hybrid methods in large sparse nonlinear least squares
    J. Optimization Theory and Applications 89, pp. 575-595.

    SIF input: Nick Gould, June 1997.

    Classification: SUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Parameters
    N: int = 100  # Number of variables (default)

    @property
    def n(self):
        """Number of variables."""
        return self.N

    @property
    def y0(self):
        """Initial guess."""
        # X(odd indices) = -1.2, X(even indices) = 1.0
        y0 = jnp.zeros(self.n)
        # Set odd indices (0, 2, 4, ...) to -1.2
        y0 = y0.at[::2].set(-1.2)
        # Set even indices (1, 3, 5, ...) to 1.0
        y0 = y0.at[1::2].set(1.0)
        return y0

    @property
    def args(self):
        """No additional arguments."""
        return None

    def objective(self, y, args):
        """Compute the least squares objective function.

        The objective is the sum of squares of M = 2*N-2 equations:
        - E(1) = X(1) - 1.0 (note: pycutest uses -1.0, SIF file shows +1.0)
        - For k=2 to 2N-3 step 2, i = (k+1)/2:
          - E(k) = 10.0 * X(i)^2 - 10.0*X(i+1)
          - E(k+1) = 2*exp(-(X(i)-X(i+1))^2) + exp(-2*(X(i+1)-X(i+2))^2)
            (note: pycutest uses + between exponentials, SIF shows -)
        - E(2N-2) = 10.0 * X(N-1)^2 (note: pycutest omits the -10.0*X(N) term)
        """
        del args  # Not used

        n = self.N

        # Vectorized computation
        # E(1) = X(1) - 1.0 (note: pycutest uses -1.0, SIF file shows +1.0)
        e1 = y[0] - 1.0

        # For the middle equations, we have pairs of equations
        # i ranges from 1 to n-2 (0-based: 0 to n-3)
        i_indices = jnp.arange(n - 2)

        # Extract variables for vectorized operations
        xi = y[i_indices]  # X(i) for i=1 to n-2
        xi1 = y[i_indices + 1]  # X(i+1)
        xi2 = y[i_indices + 2]  # X(i+2)

        # E(k) equations: 10.0 * X(i)^2 - 10.0*X(i+1)
        ek_vals = 10.0 * xi**2 - 10.0 * xi1

        # Vectorized computation of E(k+1)
        # E(k+1) = 2*exp(-(X(i)-X(i+1))^2) + exp(-2*(X(i+1)-X(i+2))^2)
        # Note: pycutest uses + between exponentials
        term1 = 2.0 * jnp.exp(-((xi - xi1) ** 2))
        term2 = jnp.exp(-2.0 * ((xi1 - xi2) ** 2))
        ek1_vals = term1 + term2

        # E(2N-2) = 10.0 * X(N-1)^2
        e_final = 10.0 * y[n - 2] ** 2

        # Combine all equations
        # We need to interleave ek_vals and ek1_vals
        # Create array with space for all equations
        equations = jnp.zeros(2 * n - 2)

        # Set E(1)
        equations = equations.at[0].set(e1)

        # Set the middle equations by interleaving
        # E(2), E(4), E(6), ... are ek_vals[0], ek_vals[1], ...
        # E(3), E(5), E(7), ... are ek1_vals[0], ek1_vals[1], ...
        equations = equations.at[1 : 2 * n - 3 : 2].set(ek_vals)
        equations = equations.at[2 : 2 * n - 3 : 2].set(ek1_vals)

        # Set E(2N-2)
        equations = equations.at[-1].set(e_final)

        # Sum of squares (L2 group type in SIF)
        # Note: The objective is (1/2) * sum of squares as per the paper
        return 0.5 * jnp.sum(equations**2)

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (not provided in SIF)."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class CSFI1(AbstractConstrainedMinimisation):
    """CSFI1 problem - continuous caster product dimensions optimization.

    Source: problem MAXTPH in Vasko and Stott
    "Optimizing continuous caster product dimensions:
     an example of a nonlinear design problem in the steel industry"
    SIAM Review, Vol 37 No, 1 pp.82-84, 1995

    SIF input: A.R. Conn April 1995

    Classification: LOR2-RN-5-4
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 5

    def objective(self, y, args):
        """Compute the objective (-TPH to maximize TPH)."""
        del args

        # Extract variables
        thick, wid, len_, tph, ipm = y

        # Objective is -TPH (minimizing -TPH = maximizing TPH)
        return -tph

    def constraint(self, y):
        """Compute the constraints separated into equalities and inequalities."""
        # Extract variables
        thick, wid, len_, tph, ipm = y

        # Parameters
        maxaspr = 2.0
        minarea = 200.0

        # CMPLQ element: 117.3708920187793427 * V1 / (V2 * V3)
        cmplq_const = 117.3708920187793427
        e1 = cmplq_const * tph / (wid * thick)

        # SQQUT element: V1 * (V1 * V2 / 48.0)
        e2 = thick * (thick * ipm / 48.0)

        # QUOTN element: V1 / V2
        e3 = wid / thick

        # PROD element: V1 * V2
        e4 = thick * wid

        # Equality constraints:
        # 1. CIPM: E1 - IPM = 0
        c1 = e1 - ipm

        # 2. CLEN: E2 - LEN = 0
        c2 = e2 - len_

        # Inequality constraints:
        # 3. WOT: E3 <= MAXASPR means E3 - MAXASPR <= 0
        c3 = e3 - maxaspr

        # 4. TTW: MINAREA <= E4 <= MAXAREA
        # pycutest returns E4 - MINAREA for the lower bound check
        c4 = e4 - minarea

        # Return equalities and inequalities separately
        equalities = jnp.array([c1, c2])
        inequalities = jnp.array([c3, c4])

        return equalities, inequalities

    @property
    def y0(self):
        """Initial guess."""
        # From START POINT section with default 0.5
        return jnp.array([0.5, 0.5, 0.5, 0.5, 0.5])

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def bounds(self):
        """Variable bounds."""
        # Parameters
        lenmax = 60.0
        minthick = 7.0

        # Bounds from SIF file
        lower = jnp.array([minthick, 0.0, 0.0, 0.0, 0.0])
        upper = jnp.array([jnp.inf, jnp.inf, lenmax, jnp.inf, jnp.inf])

        return lower, upper

    @property
    def expected_result(self):
        """Expected optimal solution (from commented start point)."""
        return jnp.array([10.01, 20.02, 60.0, 49.1, 28.75])

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # From OBJECT BOUND section
        return jnp.array(-49.1)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class BYRDSPHR(AbstractConstrainedMinimisation):
    """
    BYRDSPHR problem.

    Minimize a linear function in the intersection of two spheres.

    Source:
    R. Byrd,
    Private communication, Chicago, 1992.

    SIF input: Ph. Toint, November 1992.

    classification LQR2-AN-3-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2, x3 = y
        # Linear objective: minimize -x1 - x2 - x3
        return -x1 - x2 - x3

    @property
    def y0(self):
        # Starting point
        return jnp.array([5.0, 0.0001, -0.0001])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # No expected result given in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value given in SIF file
        return jnp.array(-4.68330049)

    @property
    def bounds(self):
        # No variable bounds
        return None

    def constraint(self, y):
        x1, x2, x3 = y

        # Two equality constraints (sphere equations)
        # SPH1: x1^2 + x2^2 + x3^2 = 9
        # SPH2: (x1-1)^2 + x2^2 + x3^2 = 9
        eq_constraint = jnp.array(
            [
                x1**2 + x2**2 + x3**2 - 9.0,  # SPH1
                (x1 - 1.0) ** 2 + x2**2 + x3**2 - 9.0,  # SPH2
            ]
        )

        # No inequality constraints
        ineq_constraint = None

        return eq_constraint, ineq_constraint



"""ALLINITC problem.

A problem with "all in it". Intended to verify that changes to LANCELOT are safe.

Source:
N. Gould: private communication.

SIF input: Nick Gould, June 1990.

classification OOR2-AY-4-1

TODO: Human review needed
Attempts made: Fixed L2 group type application (FT groups TRIVIAL, FNT groups L2)
Suspected issues: Dimension mismatch - pycutest removes fixed variable X4
Additional resources needed: Clarification on fixed variable handling
"""

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class ALLINITC(AbstractConstrainedMinimisation):
    """ALLINITC problem implementation."""

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        """Compute the objective function."""
        del args
        x1, x2, x3, x4 = y

        # Group FT1 (TRIVIAL)
        ft1 = 0.0

        # Group FT2 (TRIVIAL) with linear term
        ft2 = x3 - 1.0

        # Group FT3 (TRIVIAL) with element FT3E1 (SQR)
        ft3 = x1 * x1

        # Group FT4 (TRIVIAL) with elements FT4E1 (SQR) and FT4E2 (SQR2)
        ft4 = x2 * x2 + (x3 + x4) * (x3 + x4)

        # Group FT5 (TRIVIAL) with elements FT56E1 (SINSQR) and FT5E2 (PRODSQR)
        sinx3 = jnp.sin(x3)
        ft5 = (x4 - 3.0) + sinx3 * sinx3 + (x1 * x2) ** 2

        # Group FT6 (TRIVIAL) with element FT56E1 (SINSQR)
        ft6 = sinx3 * sinx3

        # Group FNT1 (L2)
        fnt1 = 0.0
        fnt1_l2 = fnt1 * fnt1

        # Group FNT2 (L2) with linear term
        fnt2 = x4 - 1.0
        fnt2_l2 = fnt2 * fnt2

        # Group FNT3 (L2) with element FNT3E1 (SQR)
        fnt3 = x2 * x2
        fnt3_l2 = fnt3 * fnt3

        # Group FNT4 (L2) with elements FNT4E1 (SQR) and FNT4E2 (SQR2)
        fnt4 = x3 * x3 + (x4 + x1) * (x4 + x1)
        fnt4_l2 = fnt4 * fnt4

        # Group FNT5 (L2) with elements FNT56E1 (SINSQR) and FNT5E2 (PRODSQR)
        sinx4 = jnp.sin(x4)
        fnt5 = (x1 - 4.0) + sinx4 * sinx4 + (x2 * x3) ** 2
        fnt5_l2 = fnt5 * fnt5

        # Group FNT6 (L2) with element FNT56E1 (SINSQR)
        fnt6 = sinx4 * sinx4
        fnt6_l2 = fnt6 * fnt6

        # Sum groups - FT groups are TRIVIAL (not squared), FNT groups are L2
        obj = (
            ft1
            + ft2
            + ft3
            + ft4
            + ft5
            + ft6
            + fnt1_l2
            + fnt2_l2
            + fnt3_l2
            + fnt4_l2
            + fnt5_l2
            + fnt6_l2
        )

        return obj

    def constraint(self, y):
        """Compute the constraints."""
        x1, x2 = y[0], y[1]

        # E  C1        FT3E1                    FT4E1
        # Group C1 (TRIVIAL) with elements FT3E1 (SQR) and FT4E1 (SQR)
        # Equality constraint: x1^2 + x2^2 - 1.0 = 0
        c1 = x1 * x1 + x2 * x2 - 1.0

        # Return (equality_constraints, inequality_constraints)
        # This problem has 1 equality constraint and no inequality constraints
        return jnp.array([c1]), None

    @property
    def bounds(self):
        """Return the bounds on variables."""
        # Variable bounds:
        # FR ALLINITC  X1
        # LO ALLINITC  X2        1.0
        # LO ALLINITC  X3        -1.0D+20
        # UP ALLINITC  X3        1.0
        # FX ALLINITC  X4        2.0

        y_lwr = jnp.array([-jnp.inf, 1.0, -jnp.inf, 2.0])
        y_upr = jnp.array([jnp.inf, jnp.inf, 1.0, 2.0])

        return y_lwr, y_upr

    @property
    def y0(self):
        """Return the initial point."""
        # X4 is fixed at 2.0, so we need to respect that in the initial guess
        return jnp.array([0.0, 0.0, 0.0, 2.0])

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class PALMER5E(AbstractBoundedMinimisation):
    """A nonlinear least squares problem arising from chemical kinetics.

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 T_0 + A2 T_2 + A4 T_4 + A6 T_6 + A8 T_8 +
                A10 T_10 + A12 T_12 + A14 T_14
                + L * EXP( -K X**2 )
    where T_i is the i-th (shifted) Chebyshev polynomial

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1992.

    classification SBR2-RN-8-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Data points (indices 12-23 in original data)
    M: int = 12

    # X values (radians) - starting from index 12
    X_data = jnp.array(
        [
            0.000000,
            1.570796,
            1.396263,
            1.308997,
            1.221730,
            1.125835,
            1.047198,
            0.872665,
            0.698132,
            0.523599,
            0.349066,
            0.174533,
        ]
    )

    # Y values (KJmol-1)
    Y_data = jnp.array(
        [
            4.419446,
            95.021015,
            71.6363,
            55.3397,
            41.4459,
            30.131611,
            22.2144,
            10.4570,
            3.8254,
            0.92426,
            0.06596,
            0.00000,
        ]
    )

    @property
    def n(self):
        """Number of variables."""
        return 10  # A0, A2, A4, A6, A8, A10, A12, A14, K, L

    def objective(self, y, args):
        """Compute the objective function (least squares).

        TODO: This implementation uses simple polynomials instead of Chebyshev.
        Needs proper Chebyshev polynomial calculation.
        """
        del args

        # Extract variables
        A0, A2, A4, A6, A8, A10, A12, A14, K, L = y

        # Precompute powers of X (placeholder - should use Chebyshev)
        X_sqr = self.X_data * self.X_data
        X_4 = X_sqr * X_sqr
        X_6 = X_sqr * X_4
        X_8 = X_sqr * X_6
        X_10 = X_sqr * X_8
        X_12 = X_sqr * X_10
        X_14 = X_sqr * X_12

        # Model predictions (placeholder - should use Chebyshev polynomials)
        predictions = (
            A0
            + A2 * X_sqr
            + A4 * X_4
            + A6 * X_6
            + A8 * X_8
            + A10 * X_10
            + A12 * X_12
            + A14 * X_14
            + L * jnp.exp(-K * X_sqr)
        )

        # Residuals
        residuals = predictions - self.Y_data

        # Sum of squares
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        """Initial guess."""
        return jnp.ones(10)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return jnp.zeros(10)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(2.53004766e-01)

    @property
    def bounds(self):
        """Returns the bounds on the variable y."""
        # All variables are free, but K should be non-negative for exponential decay
        lower = jnp.array(
            [
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                -jnp.inf,
                0.0,
                -jnp.inf,
            ]
        )
        upper = jnp.full(10, jnp.inf)
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class KISSING(AbstractConstrainedMinimisation):
    """KISSING NUMBER PROBLEM

    This problem is associated to the family of Hard-Spheres problem. It
    belongs to the family of sphere packing problems, a class of challenging
    problems dating from the beginning of the 17th century which is related
    to practical problems in Chemistry, Biology and Physics. It consists on
    maximizing the minimum pairwise distance between NP points on a sphere
    in R^{MDIM}.

    This problem may be reduced to a nonconvex nonlinear optimization problem
    with a potentially large number of (nonoptimal) points satisfying
    optimality conditions.

    After some algebraic manipulations, we can formulate this problem as:

                            Minimize z
                            subject to

      z >= <x_i, x_j> for all different pair of indices i, j

                            ||x_i||^2 = 1    for all i = 1,...,NP

     The goal is to find an objective value less than 0.5 (This means
     that the NP points stored belong to the sphere and every distance
     between two of them is greater than 1.0).

    References:
    [1] "Validation of an Augmented Lagrangian algorithm with a
         Gauss-Newton Hessian approximation using a set of
         Hard-Spheres problems", N. Krejic, J. M. Martinez, M. Mello
         and E. A. Pilotta, Tech. Report RP 29/98, IMECC-UNICAMP,
         Campinas, 1998.
    [2] "Inexact-Restoration Algorithm for Constrained Optimization",
         J. M. Martinez and E. A. Pilotta, Tech. Report, IMECC-UNICAMP,
         Campinas, 1998.
    [3]  "Sphere Packings, Lattices and Groups", J. H. Conway and
          N. J. C. Sloane, Springer-Verlag, NY, 1988.

    SIF input: September 29, 1998
               Jose Mario Martinez
               Elvio Angel Pilotta

    classification LQR2-RN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters
    NP = 42  # Number of points
    MDIM = 3  # Dimension

    def objective(self, y, args):
        """Objective function: minimize z"""
        # z is the last variable
        z = y[-1]
        return z

    @property
    def y0(self):
        """Starting point from SIF file"""
        # Initialize points with the provided starting values for first 12 points
        x = jnp.zeros((self.NP, self.MDIM))

        # First 12 points from SIF file
        x = x.at[0, :].set(jnp.array([-0.10890604, 0.85395078, -0.45461680]))
        x = x.at[1, :].set(jnp.array([0.49883922, -0.18439316, -0.04798594]))
        x = x.at[2, :].set(jnp.array([0.28262888, -0.48054070, 0.46715332]))
        x = x.at[3, :].set(jnp.array([-0.00580106, -0.49987584, -0.44130302]))
        x = x.at[4, :].set(jnp.array([0.81712540, -0.36874258, -0.68321896]))
        x = x.at[5, :].set(jnp.array([0.29642426, 0.82315508, 0.35938150]))
        x = x.at[6, :].set(jnp.array([0.09215152, -0.53564686, 0.00191436]))
        x = x.at[7, :].set(jnp.array([0.11700318, 0.96722760, -0.14916438]))
        x = x.at[8, :].set(jnp.array([0.01791524, 0.17759446, -0.61875872]))
        x = x.at[9, :].set(jnp.array([-0.63833630, 0.80830972, 0.45846734]))
        x = x.at[10, :].set(jnp.array([0.28446456, 0.45686938, 0.16368980]))
        x = x.at[11, :].set(jnp.array([0.76557382, 0.16700944, -0.31647534]))

        # Remaining points default to zero (as in SIF file)
        # They are already initialized to zero above

        # Flatten x and append initial z = 0.0 (default value)
        y_flat = jnp.concatenate([x.flatten(), jnp.array([0.0])])

        return y_flat

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The target is to achieve z < 0.5
        return None

    @property
    def expected_objective_value(self):
        # Should be less than 0.5
        return jnp.array(0.447214)  # Approximately sqrt(2) - 1

    @property
    def bounds(self):
        # All variables are free (XR in SIF file means free/unconstrained)
        return None

    def constraint(self, y):
        """Constraints: equality (sphere) and inequality (pairwise distances)"""
        # Extract variables
        x_flat = y[:-1]  # All x variables
        z = y[-1]  # z variable

        # Reshape x into (NP, MDIM)
        x = x_flat.reshape((self.NP, self.MDIM))

        # Equality constraints: ||x_i||^2 = 1 for all i
        equality_constraints = jnp.sum(x**2, axis=1) - 1.0

        # Inequality constraints: z >= <x_i, x_j> for all different pairs i, j
        # From SIF: XL IC(I,J) Z -1.0 means IC(I,J) + Z >= 0
        # where IC(I,J) is the dot product <x_i, x_j>
        inequality_constraints = []
        for i in range(self.NP - 1):
            for j in range(i + 1, self.NP):
                # Inner product <x_i, x_j>
                dot_product = jnp.dot(x[i], x[j])
                # Constraint: dot_product - z >= 0 (from SIF formulation)
                inequality_constraints.append(dot_product - z)

        inequality_constraints = jnp.array(inequality_constraints)

        return equality_constraints, inequality_constraints



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class HADAMALS(AbstractBoundedMinimisation):
    """Hadamard matrix search problem with ±1 entries.

    An attempt to find Hadamard matrices of order N.
    The problem is to find an N by N orthonormal matrix Q,
    with column norms N, whose entries are plus or minus one.

    Variables: N×N matrix Q flattened to N² variables
    Bounds: All variables in [-1, 1]

    Objective: Sum of squared deviations from orthogonality constraints
               plus penalties for entries not being ±1

    Source: A suggestion by Alan Edelman (MIT).
    SIF input: Nick Gould, Nov 1993.
    Classification: OBR2-RN-V-V

    Default N = 20 (400 variables).
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 20
    n_squared: int = 400

    def __init__(self, n: int = 20):
        """Initialize HADAMALS problem.

        Args:
            n: Size of the Hadamard matrix (default 20)
        """
        object.__setattr__(self, "n", n)
        object.__setattr__(self, "n_squared", n * n)

    def objective(self, y, args):
        """Compute objective function.

        The objective has two types of groups:
        1. O(I,J): Orthogonality constraints with L2 group type
        2. S(I,J): Entry constraints with LARGEL2 group type (scaled L2)
        """
        n = self.n
        rn = float(n)  # RN in SIF is N, not sqrt(N)

        # Reshape flat vector to matrix (column-major for SIF compatibility)
        Q = y.reshape((n, n), order="F")

        # Compute orthogonality terms:
        # sum over (i,j) with i<=j of (QtQ[i,j] - RN*delta[i,j])^2
        QtQ = jnp.dot(Q.T, Q)
        obj = 0.0

        # O(I,J) groups: orthogonality constraints
        for j in range(n):
            for i in range(j + 1):
                if i == j:
                    # Diagonal: QtQ[i,i] - RN where RN = N
                    val = QtQ[i, j] - rn
                else:
                    # Off-diagonal: QtQ[i,j]
                    val = QtQ[i, j]
                obj = obj + val * val

        # S(I,J) groups: entry constraints (Q[i,j]^2 - 1)^2 for i>=2
        # Using LARGEL2 group type with FACTOR=1.0
        for j in range(n):
            for i in range(1, n):  # i from 2 to N (0-indexed: 1 to n-1)
                val = Q[i, j] * Q[i, j] - 1.0
                obj = obj + val * val

        return jnp.array(obj)

    @property
    def y0(self):
        """Starting point as specified in SIF file."""
        n = self.n
        n_half = n // 2

        # Initialize matrix
        Q0 = jnp.zeros((n, n))

        # Set initial values for each column as per SIF file
        # Note: The bounds will be enforced separately, so we don't override here
        for j in range(n):
            # First half of rows: 0.9
            Q0 = Q0.at[:n_half, j].set(0.9)
            # Second half of rows: -0.9
            Q0 = Q0.at[n_half:, j].set(-0.9)

        # Flatten in column-major order (Fortran convention)
        return Q0.flatten(order="F")

    @property
    def bounds(self):
        """All variables bounded in [-1, 1], with first column fixed."""
        n = self.n
        n_half = n // 2

        # Default bounds: [-1, 1] for all variables
        lower = -jnp.ones(self.n_squared)
        upper = jnp.ones(self.n_squared)

        # Fix first column values (column-major indexing)
        # Q(i,1) in SIF is at index i in column-major flattened array
        for i in range(n_half):
            lower = lower.at[i].set(1.0)
            upper = upper.at[i].set(1.0)
        for i in range(n_half, n):
            lower = lower.at[i].set(-1.0)
            upper = upper.at[i].set(-1.0)

        return lower, upper

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Optimal solution would be a true Hadamard matrix if one exists
        return None

    @property
    def expected_objective_value(self):
        # Perfect Hadamard matrix would have objective 0
        return jnp.array(0.0)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class LUKSAN11(AbstractNonlinearEquations):
    """
    Problem 11 (chained serpentine) in the paper

    L. Luksan
    Hybrid methods in large sparse nonlinear least squares
    J. Optimization Theory & Applications 89(3) 575-595 (1996)

    SIF input: Nick Gould, June 2017.

    classification NOR2-AN-V-V
    """

    s: int = 99  # Seed for dimensions
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, s: int = 99):
        self.s = s

    @property
    def n(self) -> int:
        """Number of variables: S + 1."""
        return self.s + 1

    def num_residuals(self) -> int:
        """Number of residuals: 2 * S."""
        return 2 * self.s

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        return jnp.full(self.n, -0.8, dtype=jnp.float64)

    def residual(self, y: Array, args) -> Array:
        """Compute the residual vector."""
        x = y
        s = self.s

        # Vectorized computation
        # Extract x[0:s] and x[1:s+1] for vectorized operations
        x_i = x[:-1]  # x[0] to x[s-1]
        x_ip1 = x[1:]  # x[1] to x[s]

        # First residuals: 20*x[i]/(1+x[i]^2) - 10*x[i+1]
        d = 1.0 + x_i * x_i
        res1 = 20.0 * x_i / d - 10.0 * x_ip1

        # Second residuals: x[i] with RHS adjustment
        res2 = x_i.copy()

        # Create residuals array by interleaving res1 and res2
        residuals = jnp.zeros(2 * s, dtype=jnp.float64)
        residuals = residuals.at[::2].set(res1)  # Even indices: 0, 2, 4, ...
        residuals = residuals.at[1::2].set(res2)  # Odd indices: 1, 3, 5, ...

        # Apply constants: subtract 1.0 from odd indices (1, 3, 5, ...)
        # This corresponds to RHS = 1.0 for alternating equations
        odd_mask = jnp.arange(2 * s) % 2 == 1
        residuals = residuals - odd_mask.astype(jnp.float64)

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Solution should satisfy residuals = 0
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """Free bounds for all variables."""
        return None



from jax import numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractNonlinearEquations


class HATFLDBNE(AbstractNonlinearEquations):
    """A test problem from the OPTIMA user manual.

    Source:
    "The OPTIMA user manual (issue No.8, p. 12)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.
    Bound-constrained nonlinear equations version: Nick Gould, June 2019.

    classification NOR2-AN-4-4
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def residual(self, y, args) -> Float[Array, "4"]:
        """Residual function for the nonlinear equations."""
        x = y

        # G(1): X(1) = 1.0
        res1 = x[0] - 1.0

        # G(i) for i=2..4: X(i-1) - sqrt(X(i)) = 0
        residuals = [res1]
        for i in range(1, 4):  # i = 2, 3, 4 in 1-indexed notation
            residuals.append(x[i - 1] - jnp.sqrt(x[i]))

        return jnp.array(residuals)

    @property
    def y0(self) -> Float[Array, "4"]:
        """Initial guess for the optimization problem."""
        return jnp.array([0.1, 0.1, 0.1, 0.1])

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file doesn't provide a solution
        return None

    @property
    def expected_objective_value(self) -> Float[Array, ""]:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    @property
    def bounds(self) -> tuple[Float[Array, "4"], Float[Array, "4"]]:
        """Bounds on variables."""
        # Lower bounds: all 0.0000001
        lower = jnp.array([0.0000001, 0.0000001, 0.0000001, 0.0000001])
        # Upper bounds: only X2 has upper bound of 0.8
        upper = jnp.array([jnp.inf, 0.8, jnp.inf, jnp.inf])
        return lower, upper

    def num_constraints(self) -> tuple[int, int, int]:
        """Returns the number of constraints."""
        # 4 equality constraints (the residuals) + 5 finite bounds (4 lower + 1 upper)
        num_equalities = 4
        num_inequalities = 0
        num_bounds = 5  # 4 finite lower bounds + 1 finite upper bound
        return num_equalities, num_inequalities, num_bounds

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class VARDIMNE(AbstractNonlinearEquations):
    """Variable dimension nonlinear equations problem.

    This problem is a sum of n+2 least-squares groups, the first n of
    which have only a linear element. Its Hessian matrix is dense.
    This is a nonlinear equation version of problem VARDIM.

    Source: problem 25 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#72 (p.98).

    SIF input: Ph. Toint, Dec 1989.
    Modification as a set of nonlinear equations: Nick Gould, Oct 2015.

    Classification: NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem dimension
    N: int = 10  # Default from SIF file

    @property
    def n(self):
        """Number of variables."""
        return self.N

    @property
    def m(self):
        """Number of equations."""
        return self.N + 2

    def residual(self, y):
        """Compute the residuals for the nonlinear equations.

        The residuals are:
        - First n equations: x_i - 1 for i=1..n
        - Equation n+1: sum_{i=1}^n i*x_i - n*(n+1)/2
        - Equation n+2: (sum_{i=1}^n i*x_i - n*(n+1)/2)^2
        """
        # Compute n*(n+1)/2
        n_times_n_plus_1_over_2 = self.N * (self.N + 1) / 2.0

        # Linear terms sum_i i*x_i
        indices = jnp.arange(1, self.N + 1, dtype=y.dtype)
        linear_sum = jnp.sum(indices * y)

        # First n residuals: x_i - 1
        first_n_residuals = y - 1.0

        # Residual n+1: sum_i i*x_i - n*(n+1)/2
        residual_n_plus_1 = linear_sum - n_times_n_plus_1_over_2

        # Residual n+2: (sum_i i*x_i - n*(n+1)/2)^2
        # Note: This is the squared version for the nonlinear equation
        residual_n_plus_2 = residual_n_plus_1**2

        # Combine all residuals
        residuals = jnp.concatenate(
            [
                first_n_residuals,
                jnp.array([residual_n_plus_1]),
                jnp.array([residual_n_plus_2]),
            ]
        )

        return residuals

    def constraint(self, y):
        """Return the equality constraints (residuals) for the nonlinear system."""
        residuals = self.residual(y)
        # Return (equality_constraints, inequality_constraints)
        # For nonlinear equations, we have only equality constraints
        return residuals, None

    @property
    def y0(self):
        """Initial guess: x_i = 1 - i/n."""
        indices = jnp.arange(1, self.N + 1, dtype=jnp.float64)
        return 1.0 - indices / self.N

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def bounds(self):
        """No bounds for this problem."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution is x_i = 1 for all i."""
        return jnp.ones(self.N)

    @property
    def expected_objective_value(self):
        """Expected optimal objective value is 0."""
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: human review required
class HIMMELBCLS(AbstractUnconstrainedMinimisation):
    """Himmelblau's nonlinear least-squares problem.

    A 2-variable problem with 2 residuals.

    Source: problem 201 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lecture Notes in Economics and Mathematical Systems 187,
    Springer Verlag, Berlin, 1981.

    SIF input: Ph. Toint, Dec 1989.
    Classification: SUR2-AN-2-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2 = y

        # From SIF file:
        # G1 = X2 + X1² - 11.0
        # G2 = X1 + X2² - 7.0
        # Objective = G1² + G2²
        g1 = x2 + x1**2 - 11.0
        g2 = x1 + x2**2 - 7.0

        # Return the sum of squares
        return g1**2 + g2**2

    @property
    def y0(self):
        return jnp.array([1.0, 1.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The exact solution is (±√7, ±√11)
        # We'll return one of the four solutions
        return jnp.array([jnp.sqrt(7.0), jnp.sqrt(11.0)])

    @property
    def expected_objective_value(self):
        # At the optimal solution, both residuals are 0
        return jnp.array(0.0)



from jax import numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractNonlinearEquations


class ENGVAL2NE(AbstractNonlinearEquations):
    """The ENGVAL2NE problem.

    Source: problem 15 in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.
    Nonlinear-equations version of ENGVAL2.SIF, Nick Gould, Jan 2020.

    classification NOR2-AN-3-5
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def residual(self, y, args) -> Float[Array, "5"]:
        """Residual function for the nonlinear equations."""
        x1, x2, x3 = y

        # G1: X1^2 + X2^2 + X3^2 = 1.0
        res1 = x1**2 + x2**2 + x3**2 - 1.0

        # G2: X1^2 + X2^2 + (X3-2)^2 = 1.0
        res2 = x1**2 + x2**2 + (x3 - 2.0) ** 2 - 1.0

        # G3: X1 + X2 + X3 = 1.0
        res3 = x1 + x2 + x3 - 1.0

        # G4: X1 + X2 - X3 = -1.0
        res4 = x1 + x2 - x3 + 1.0

        # G5: 3*X2^2 + X1^3 + (5*X3 - X1 + 1)^2 = 36.0
        twow = 5.0 * x3 - x1 + 1.0
        res5 = 3.0 * x2**2 + x1**3 + twow**2 - 36.0

        return jnp.array([res1, res2, res3, res4, res5])

    @property
    def y0(self) -> Float[Array, "3"]:
        """Initial guess for the optimization problem."""
        return jnp.array([1.0, 2.0, 0.0])

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file doesn't provide a solution
        return None

    @property
    def expected_objective_value(self) -> Float[Array, ""]:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractConstrainedMinimisation


class ORTHREGC(AbstractConstrainedMinimisation):
    """
    An orthogonal regression problem.

    The problem is to fit (orthogonally) an ellipse to a set of points
    in the plane. This set of points is generated by perturbing a
    first set lying exactly on a predefined ellipse centered at the
    origin.

    Source: adapted from:
    M. Gulliksson,
    "Algorithms for nonlinear Least-squares with Applications to
    Orthogonal Regression",
    UMINF-178.90, University of Umea, Sweden, 1990.

    SIF input: Ph. Toint, June 1990.

    classification QQR2-AN-V-V

    Number of data points (number of variables = 2 NPTS + 5)
    Default NPTS = 2500
    """

    npts: int = 2500  # Number of data points (default from SIF)
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, npts: int = 2500):
        self.npts = npts

    @property
    def n(self) -> int:
        """Total number of variables: 5 ellipse parameters + 2 * NPTS projections."""
        return 5 + 2 * self.npts

    @property
    def m(self) -> int:
        """Number of constraints: NPTS equality constraints."""
        return self.npts

    def _generate_data_points(self):
        """Generate the perturbed ellipse data points."""
        # True ellipse parameters
        v11, v12, v22 = 2.0, 1.0, 2.0

        # Perturbation parameters
        pseed = 237.1531
        psize = 0.2
        pi = jnp.pi

        # Computed parameters
        incr = 2.0 * pi / self.npts
        c3 = jnp.cos(v22)
        s3 = jnp.sin(v22)

        # Generate angles
        i_vals = jnp.arange(self.npts, dtype=jnp.float64)
        theta = i_vals * incr

        # Compute sin and cos
        st = jnp.sin(theta)
        ct = jnp.cos(theta)

        # Ellipse points before rotation
        u1 = v11 * ct
        u2 = v12 * st

        # Apply rotation by angle v22
        r1 = u1 * c3 - u2 * s3
        r2 = u1 * s3 + u2 * c3

        # Add perturbation
        xseed = theta * pseed
        sseed = jnp.cos(xseed)
        pert = 1.0 + psize * sseed

        # Data points
        xd = r1 * pert
        yd = r2 * pert

        return xd, yd

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        y = jnp.zeros(self.n, dtype=jnp.float64)

        # Ellipse parameters (H11, H12, H22, G1, G2)
        y = y.at[0].set(1.0)  # H11
        y = y.at[1].set(0.0)  # H12
        y = y.at[2].set(1.0)  # H22
        y = y.at[3].set(1.0)  # G1
        y = y.at[4].set(1.0)  # G2

        # Point projections initialized to data points (interleaved X,Y)
        xd, yd = self._generate_data_points()
        y = y.at[5::2].set(xd)  # Set all X values at indices 5, 7, 9, ...
        y = y.at[6::2].set(yd)  # Set all Y values at indices 6, 8, 10, ...

        return y

    def objective(self, y: Array, args) -> Array:
        """Compute the objective function."""
        # Extract projected points (interleaved X,Y)
        x_proj = y[5::2]  # Gets indices 5, 7, 9, ... (all X values)
        y_proj = y[6::2]  # Gets indices 6, 8, 10, ... (all Y values)

        # Get data points
        xd, yd = self._generate_data_points()

        # Sum of squared distances to data points
        obj = jnp.sum((x_proj - xd) ** 2 + (y_proj - yd) ** 2)

        return obj

    def constraint(self, y: Array):
        """Compute the equality constraints."""
        # Extract ellipse parameters
        h11, h12, h22, g1, g2 = y[0], y[1], y[2], y[3], y[4]

        # Extract projected points (interleaved X,Y)
        x_proj = y[5::2]  # Gets indices 5, 7, 9, ... (all X values)
        y_proj = y[6::2]  # Gets indices 6, 8, 10, ... (all Y values)

        # Ellipse constraints: H11*x^2 + 2*H12*x*y + H22*y^2 - 2*G1*x - 2*G2*y - 1 = 0
        eq_constraints = (
            h11 * x_proj**2
            + 2.0 * h12 * x_proj * y_proj
            + h22 * y_proj**2
            - 2.0 * g1 * x_proj
            - 2.0 * g2 * y_proj
            - 1.0
        )

        ineq_constraints = None

        return eq_constraints, ineq_constraints

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """All variables are free."""
        return None

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the objective and constraint functions."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Not explicitly given in the SIF file
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # Different values for different NPTS sizes given in SIF comments
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractUnconstrainedMinimisation


class PENALTY2(AbstractUnconstrainedMinimisation):
    """
    PENALTY2 problem.

    The second penalty function

    This is a nonlinear least-squares problem with M=2*N groups.
     Group 1 is linear.
     Groups 2 to N use 2 nonlinear elements.
     Groups N+1 to M-1 use 1 nonlinear element.
     Group M uses N nonlinear elements.
    The Hessian matrix is dense.

    Source:  Problem 24 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#112 (p. 80)

    SIF input: Ph. Toint, Dec 1989.

    classification SUR2-AN-V-0

    TODO: Human review needed
    Attempts made: Multiple interpretations of SCALE factor in SIF
    Suspected issues: Incorrect understanding of how SCALE interacts with L2 groups
    The objective/gradient values are off by a factor of ~10^10
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 200

    def objective(self, y, args):
        del args
        n = self.n

        # Parameters
        a = 0.00001
        b = 1.0

        # Group 1: (x[0] - 0.2)^2 scaled by 1/b
        # Note: SCALE in SIF might multiply the squared value
        g1 = (y[0] - 0.2) ** 2 / b

        # Precompute exp(0.1*x) for all elements
        exp_x = jnp.exp(0.1 * y)

        # Groups 2 to n: 1/a * (exp(0.1*x[i]) + exp(0.1*x[i-1]) -
        #                        (exp(0.1*i) + exp(0.1*(i-1))))^2
        if n > 1:
            # Vectorized computation
            i_vals = jnp.arange(2, n + 1)  # i from 2 to n
            y_vals = jnp.exp(0.1 * i_vals) + jnp.exp(0.1 * (i_vals - 1))
            a_vals = exp_x[1:n]  # exp(0.1*x[i]) for i=2 to n (0-based: 1 to n-1)
            b_vals = exp_x[
                0 : n - 1
            ]  # exp(0.1*x[i-1]) for i=2 to n (0-based: 0 to n-2)

            group_vals = a_vals + b_vals - y_vals
            groups_2_to_n = (1.0 / a) * jnp.sum(group_vals**2)
        else:
            groups_2_to_n = 0.0

        # Groups n+1 to m-1: 1/a * (exp(0.1*x[i-n+1]) - exp(-0.1))^2
        if n > 1:
            # i from n+1 to 2n-1, so i-n+1 from 2 to n (0-based: 1 to n-1)
            c_vals = exp_x[1:n]  # exp(0.1*x[j]) for j=2 to n
            em1_10 = jnp.exp(-0.1)
            group_vals = c_vals - em1_10
            groups_np1_to_m1 = (1.0 / a) * jnp.sum(group_vals**2)
        else:
            groups_np1_to_m1 = 0.0

        # Group m: 1/b * (sum_{j=1}^n (n-j+1)*x[j]^2 - 1)^2
        # Weights: n, n-1, ..., 1
        weights = jnp.arange(n, 0, -1)
        weighted_sum = jnp.sum(weights * y**2)
        gm = (1.0 / b) * (weighted_sum - 1.0) ** 2

        return g1 + groups_2_to_n + groups_np1_to_m1 + gm

    @property
    def y0(self):
        # Standard starting point (0.5, ..., 0.5)
        return inexact_asarray(0.5 * jnp.ones(self.n))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution values provided for some values of n
        n = self.n
        if n == 4:
            return jnp.array(9.37629e-6)
        elif n == 10:
            return jnp.array(2.93660e-4)
        elif n == 50:
            return jnp.array(4.29609813)
        elif n == 100:
            return jnp.array(97096.0840)
        else:
            # For n=200 and others, solution not provided
            return None

    def num_variables(self):
        return self.n



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class CANTILVR(AbstractConstrainedMinimisation):
    """
    CANTILVR problem.

    Computation of a minimum weight cantilever consisting of 5 sections of
    square shaped tube of given thickness.

    Source:
    an example in a talk by W.K. Zhang and C. Fleury, LLN, 1994.

    SIF input: Ph. Toint, November 1994
               correction by S. Gratton & Ph. Toint, May 2024

    classification LOR2-MN-5-1
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2, x3, x4, x5 = y
        # Linear objective: minimize weight
        return 0.0624 * (x1 + x2 + x3 + x4 + x5)

    @property
    def y0(self):
        # Starting point
        return jnp.ones(5)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # No expected result given in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value given in SIF file
        return jnp.array(1.3399595938)

    @property
    def bounds(self):
        # Variable bounds: 0.000001 <= x <= inf
        lower = jnp.full(5, 0.000001)
        upper = jnp.full(5, jnp.inf)
        return (lower, upper)

    def constraint(self, y):
        x1, x2, x3, x4, x5 = y

        # No equality constraints
        eq_constraint = None

        # Single inequality constraint:
        # 61/x1^3 + 37/x2^3 + 19/x3^3 + 7/x4^3 + 1/x5^3 <= 1
        # NOTE: This constraint is undefined at x=0 due to division by x^3
        # This causes expected test failures when evaluating at zero vector
        # In pycutest format (L-type), this is returned as:
        # constraint_value - constant
        ineq_constraint = jnp.array(
            [
                61.0 / x1**3
                + 37.0 / x2**3
                + 19.0 / x3**3
                + 7.0 / x4**3
                + 1.0 / x5**3
                - 1.0
            ]
        )

        return eq_constraint, ineq_constraint



"""Compatibility wrapper for converting between constraint conventions (DRAFT).

This module provides a wrapper class that converts constraints between different
sign conventions used by PyCUTEst and Optimistix:

- PyCUTEst: Uses mixed conventions (g(x) >= 0, g(x) <= 0, and range constraints)
- Optimistix: Uses consistent g(x) >= 0 convention for all inequality constraints

The wrapper allows problems implemented in Optimistix convention to be tested
against PyCUTEst reference implementations by converting constraints appropriately.
"""

from typing import Any

from jaxtyping import Array, ArrayLike

from sif2jax import AbstractConstrainedMinimisation


class PyCUTEstCompatibilityWrapper(AbstractConstrainedMinimisation):
    """Wrapper that converts Optimistix-convention constraints to PyCUTEst format.

    This wrapper takes a problem implemented with Optimistix's g(x) >= 0 convention
    and converts its constraints to match PyCUTEst's mixed convention for testing.

    The wrapper:
    1. Passes through all problem methods unchanged except constraint()
    2. Converts inequality constraints based on their types:
       - 'G' constraints: passed through (already g(x) >= 0)
       - 'L' constraints: negated (from -g(x) >= 0 to g(x) <= 0)
       - 'R' constraints: reconstructed from two inequalities to single ranged form
       - 'E' constraints: passed through unchanged

    Constraint type information is expected to be available as attributes on the
    wrapped problem instance.

    Attributes:
        problem: The underlying problem instance with Optimistix conventions
    """

    problem: AbstractConstrainedMinimisation

    @property
    def y0_iD(self) -> int:
        """Pass through y0_iD from wrapped problem."""
        return self.problem.y0_iD

    @property
    def provided_y0s(self) -> frozenset:
        """Pass through provided_y0s from wrapped problem."""
        return self.problem.provided_y0s

    # Pass through all methods to the underlying problem
    def objective(self, y: ArrayLike, args: Any) -> Array:
        return self.problem.objective(y, args)

    @property
    def y0(self) -> Array:
        return self.problem.y0

    @property
    def args(self) -> Any:
        return self.problem.args

    @property
    def expected_result(self) -> Array | None:
        return self.problem.expected_result

    @property
    def expected_objective_value(self) -> Array | None:
        return self.problem.expected_objective_value

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        return self.problem.bounds

    @property
    def name(self) -> str:
        return self.problem.name

    def constraint(self, y: ArrayLike) -> tuple[Array, Array]:
        """Convert constraints from Optimistix to PyCUTEst format.

        This method transforms constraints from consistent g(x) >= 0 format
        to PyCUTEst's mixed format based on constraint_types metadata.

        Args:
            y: Point at which to evaluate constraints

        Returns:
            Tuple of (equality_constraints, inequality_constraints) in PyCUTEst format

        Raises:
            NotImplementedError: Always (to be implemented based on feedback)
        """
        raise NotImplementedError("Constraint conversion logic to be implemented")



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class BRANIN(AbstractBoundedMinimisation):
    """The Branin & Ho function - a global optimization test problem.

    The Branin function has the form:
    a(x2 - b*x1**2 + c*x1 - r)**2 + s(1-t)cos(x1) + s
    where a=1, b=5.1/(4π²), c=5/π, s=10 and t=1/(8π),
    and x1 in [-5, 10], x2 in [0, 15]

    Source:
    L. C. W. Dixon and G. P. Szego (Eds.)
    Towards Global Optimization
    North Holland, 1975.

    SIF input: Nick Gould, July 2021

    classification: OBR2-AN-2-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 2  # 2 variables

    @property
    def y0(self):
        # Starting point from SIF file
        return jnp.array([5.0, 10.0])

    @property
    def args(self):
        return ()

    def objective(self, y, args):
        """Compute the objective function.

        Minimize: a(x2 - b*x1**2 + c*x1 - r)**2 + s(1-t)cos(x1) + s
        """
        x1, x2 = y

        # Parameters
        pi = jnp.pi
        a = 1.0
        b = 5.1 / (4.0 * pi**2)
        c = 5.0 / pi
        r = 6.0
        s = 10.0
        t = 1.0 / (8.0 * pi)

        # First term: a(x2 - b*x1**2 + c*x1 - r)**2
        inner = x2 - b * x1**2 + c * x1 - r
        term1 = a * inner**2

        # Second term: s(1-t)cos(x1) + s
        term2 = s * (1.0 - t) * jnp.cos(x1) + s

        return term1 + term2

    @property
    def bounds(self):
        """Returns the bounds on the variables."""
        # From SIF file: x1 in [-5, 10], x2 in [0, 15]
        lower = jnp.array([-5.0, 0.0])
        upper = jnp.array([10.0, 15.0])
        return lower, upper

    @property
    def expected_result(self):
        # The Branin function has three global minima:
        # (-π, 12.275), (π, 2.275), (9.42478, 2.475)
        # Return one of them
        return jnp.array([-jnp.pi, 12.275])

    @property
    def expected_objective_value(self):
        # The global minimum value
        return jnp.array(0.397887)



# TODO: Human review needed
# Constraint dimension mismatch: pycutest returns 998 equality constraints
# but our implementation expects 499. This needs investigation to determine
# the correct constraint formulation

from jax import numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractNonlinearEquations


class ARWHDNE(AbstractNonlinearEquations):
    """A quartic problem whose Hessian is an arrow-head (downwards) with
    diagonal central part and border-width of 1.
    Nonlinear equation version of problem ARWHEAD.

    Source: Problem 55 in
    A.R. Conn, N.I.M. Gould, M. Lescrenier and Ph.L. Toint,
    "Performance of a multifrontal scheme for partially separable
    optimization",
    Report 88/4, Dept of Mathematics, FUNDP (Namur, B), 1988.

    SIF input: Ph. Toint, Dec 1989.

    classification NOR2-AN-V-V
    """

    n: int = 500  # Default to n=500
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def residual(self, y, args) -> Float[Array, "499"]:
        """Residual function for the nonlinear equations."""
        x = y
        n = self.n
        ngs = n - 1  # Number of group sets

        # Initialize residuals array
        residuals = []

        # For each group set i = 1, ..., n-1
        for i in range(ngs):
            # Linear part: -4*x(i) - 3
            linear_part = -4.0 * x[i] - 3.0

            # Nonlinear part: x(i)^2 + x(n)^2
            nonlinear_part = x[i] * x[i] + x[n - 1] * x[n - 1]

            # Combined residual for group i
            residuals.append(linear_part + nonlinear_part)

        return jnp.array(residuals)

    @property
    def y0(self) -> Float[Array, "500"]:
        """Initial guess for the optimization problem."""
        return jnp.ones(self.n)

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file doesn't provide a solution
        return None

    @property
    def expected_objective_value(self) -> Float[Array, ""]:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class NONDQUAR(AbstractUnconstrainedMinimisation):
    """
    NONDQUAR problem.

    A nondiagonal quartic test problem.

    This problem has an arrow-head type Hessian with a tridiagonal
    central part and a border of width 1.
    The Hessian is singular at the solution.

    Source: problem 57 in
    A.R. Conn, N.I.M. Gould, M. Lescrenier and Ph.L. Toint,
    "Performance of a multi-frontal scheme for partially separable
    optimization"
    Report 88/4, Dept of Mathematics, FUNDP (Namur, B), 1988.

    SIF input: Ph. Toint, Dec 1989.

    classification OUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 5000

    def objective(self, y, args):
        del args
        n = self.n

        # First n-2 groups: (x[i] + x[i+1] + x[n-1])^4 for i=0 to n-3
        if n > 2:
            # Vectorized computation
            x_i = y[0 : n - 2]
            x_ip1 = y[1 : n - 1]
            x_n = y[n - 1]
            group_values = x_i + x_ip1 + x_n
            l4_terms = group_values**4
            l4_sum = jnp.sum(l4_terms)
        else:
            l4_sum = 0.0

        # Last two groups use L2 (squared) type
        # L(N-1): (x[0] - x[1])^2
        ln_1 = (y[0] - y[1]) ** 2

        # L(N): (x[n-2] - x[n-1])^2
        ln = (y[n - 2] - y[n - 1]) ** 2

        return l4_sum + ln_1 + ln

    @property
    def y0(self):
        # Starting from (1, -1, 1, -1, ... )
        n = self.n
        y0 = jnp.ones(n)
        y0 = y0.at[1::2].set(-1.0)
        return y0

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value is 0.0
        return jnp.array(0.0)

    def num_variables(self):
        return self.n



"""Hartman 6-dimensional test problem."""

from __future__ import annotations

from typing import Any

import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class HART6(AbstractBoundedMinimisation):
    """Hartman 6-dimensional test problem.

    Source: Hartman problem 6 in
    L. C. W. Dixon and G. P. Szego (Eds.)
    Towards Global Optimization
    North Holland, 1975.
    Paper 9, page 163.

    SIF input: A.R. Conn May 1995

    classification OBR2-AN-6-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y: Any, args: Any) -> Any:
        """Compute the objective function."""
        # Coefficients c_i
        c = jnp.array([1.0, 1.2, 3.0, 3.2])

        # Matrix A (4x6)
        a = jnp.array(
            [
                [10.0, 0.05, 17.0, 3.5, 1.7, 8.0],
                [0.05, 10.0, 17.0, 0.1, 8.0, 14.0],
                [3.0, 3.5, 1.7, 10.0, 17.0, 8.0],
                [17.0, 8.0, 0.05, 10.0, 0.1, 14.0],
            ]
        )

        # Matrix P (4x6)
        p = jnp.array(
            [
                [0.1312, 0.1696, 0.5569, 0.0124, 0.8283, 0.5886],
                [0.2329, 0.4135, 0.8307, 0.3736, 0.1004, 0.9991],
                [0.2348, 0.1451, 0.3522, 0.2883, 0.3047, 0.6650],
                [0.4047, 0.8828, 0.8732, 0.5743, 0.1091, 0.0381],
            ]
        )

        # Compute objective: sum of negative exponentials
        obj = 0.0
        for i in range(4):
            # Compute sum of a[i,j] * (x[j] - p[i,j])^2
            diff = y - p[i, :]
            weighted_sum = jnp.sum(a[i, :] * diff * diff)
            obj = obj - c[i] * jnp.exp(-weighted_sum)

        return obj

    @property
    def bounds(self):
        """Variable bounds: 0 <= x_i <= 1 for all i."""
        lower = jnp.zeros(6)
        upper = jnp.ones(6)
        return lower, upper

    @property
    def y0(self):
        return jnp.full(6, 0.2)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The optimal solution from the commented section in SIF
        return jnp.array([0.201661, 0.149985, 0.476919, 0.275317, 0.311688, 0.657283])

    @property
    def expected_objective_value(self):
        # From SIF file
        return jnp.array(-3.32288689158)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class DTOC4(AbstractConstrainedMinimisation):
    """A discrete time optimal control (DTOC) problem with quadratic objective.

    The system has N time periods, 1 control variable and 2 state variables.

    The problem is not convex.

    Sources: problem 4 in
    T.F. Coleman and A. Liao,
    "An Efficient Trust Region Method for Unconstrained Discrete-Time Optimal
    Control Problems",
    Tech. Report, ctc93tr144,  Advanced Computing Research Institute,
    Cornell University, 1992.

    G. Di Pillo, L. Grippo and F. Lampariello,
    "A class of structures quasi-Newton algorithms for optimal control
    problems",
    in H.E. Rauch, ed., IFAC Applications of nonlinear programming to
    optimization and control, pp. 101-107, IFAC, Pergamon Press, 1983.

    SIF input: Ph. Toint, August 1993

    classification QOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters - using the default values from SIF
    n_periods: int = 1500  # N
    n_controls: int = 1  # NX=1
    n_states: int = 2  # NY=2

    # Total number of variables and constraints
    n: int = (n_periods - 1) * n_controls + n_periods * n_states  # 4499
    m: int = (n_periods - 1) * n_states  # 2998

    @property
    def y0(self):
        # Start from zeros except for fixed values
        y = jnp.zeros(self.n)
        # Y(1,1) = 0.0 is at index 1499
        # Y(1,2) = 1.0 is at index 1500
        y = y.at[1500].set(1.0)
        return y

    @property
    def args(self):
        # H = 1/N and related constants
        h = 1.0 / self.n_periods
        h5 = 5.0 * h
        return h, h5

    def objective(self, y, args):
        h, h5 = args
        # The SIF file says scale = "1/5H" but pycutest interprets this
        # as dividing by 5H, not multiplying by 1/(5H)
        scale_factor = h5  # 5H (reciprocal of what we had)

        # Extract control and state variables
        # Note: We don't enforce fixed variables here as pycutest evaluates
        # with the values given in the input vector
        # X(t) for t=1..N-1
        # Y(t,i) for t=1..N, i=1..2
        x_vars = y[: (self.n_periods - 1) * self.n_controls].reshape(
            self.n_periods - 1, self.n_controls
        )
        y_vars = y[(self.n_periods - 1) * self.n_controls :].reshape(
            self.n_periods, self.n_states
        )

        # Objective function has structure:
        # (1/5H) * [
        #   0.5*Y(1,1)^2 + 0.5*Y(1,2)^2 + X(1)^2 +
        #   sum_{t=2}^{N-1} (Y(t,1)^2 + Y(t,2)^2 + X(t)^2) +
        #   0.5*Y(N,1)^2 + 0.5*Y(N,2)^2
        # ]

        # First term (t=1)
        obj_sum = 0.5 * y_vars[0, 0] ** 2 + 0.5 * y_vars[0, 1] ** 2 + x_vars[0, 0] ** 2

        # Middle terms (t=2..N-1)
        if self.n_periods > 2:
            y_middle = y_vars[1:-1]  # Y(t) for t=2..N-1
            x_middle = x_vars[1:]  # X(t) for t=2..N-1
            obj_sum += jnp.sum(y_middle[:, 0] ** 2 + y_middle[:, 1] ** 2)
            obj_sum += jnp.sum(x_middle[:, 0] ** 2)

        # Last term (t=N)
        obj_sum += 0.5 * y_vars[-1, 0] ** 2 + 0.5 * y_vars[-1, 1] ** 2

        return scale_factor * obj_sum

    def constraint(self, y):
        h, h5 = self.args

        # Extract control and state variables
        # Note: We don't enforce fixed variables here as pycutest evaluates
        # with the values given in the input vector
        x_vars = y[: (self.n_periods - 1) * self.n_controls].reshape(
            self.n_periods - 1, self.n_controls
        )
        y_vars = y[(self.n_periods - 1) * self.n_controls :].reshape(
            self.n_periods, self.n_states
        )

        # Transition constraints
        # TT(t,1): -Y(t+1,1) + (1+5H)*Y(t,1) - 5H*Y(t,2) + 5H*X(t) -
        #          5H*Y(t,2)^2*Y(t,1) = 0
        # TT(t,2): -Y(t+1,2) + Y(t,2) + 5H*Y(t,1) = 0

        # Vectorize the computation
        y_current = y_vars[:-1]  # Y(t) for t=1..N-1
        y_next = y_vars[1:]  # Y(t+1) for t=1..N-1

        # Shape: (n_periods-1, n_states)
        constraints = jnp.zeros((self.n_periods - 1, self.n_states))

        # First constraint for each t: includes AAB element
        # TT(t,1) = -Y(t+1,1) + (1+5H)*Y(t,1) - 5H*Y(t,2) + 5H*X(t) - 5H*Y(t,2)^2*Y(t,1)
        aab_term = y_current[:, 1] ** 2 * y_current[:, 0]  # Y(t,2)^2 * Y(t,1)
        constraints = constraints.at[:, 0].set(
            -y_next[:, 0]
            + (1 + h5) * y_current[:, 0]
            - h5 * y_current[:, 1]
            + h5 * x_vars[:, 0]
            - h5 * aab_term
        )

        # Second constraint for each t
        # TT(t,2) = -Y(t+1,2) + Y(t,2) + 5H*Y(t,1)
        constraints = constraints.at[:, 1].set(
            -y_next[:, 1] + y_current[:, 1] + h5 * y_current[:, 0]
        )

        # Flatten to match expected output shape
        equality_constraints = constraints.flatten()

        return equality_constraints, None

    @property
    def bounds(self):
        # Y(1,1)=0, Y(1,2)=1 are fixed, others are free
        n_x = (self.n_periods - 1) * self.n_controls
        lbs = jnp.full(self.n, -jnp.inf)
        ubs = jnp.full(self.n, jnp.inf)

        # Y(1,1) at index n_x (1499) is fixed to 0.0
        lbs = lbs.at[n_x].set(0.0)
        ubs = ubs.at[n_x].set(0.0)

        # Y(1,2) at index n_x+1 (1500) is fixed to 1.0
        lbs = lbs.at[n_x + 1].set(1.0)
        ubs = ubs.at[n_x + 1].set(1.0)

        return lbs, ubs

    @property
    def expected_result(self) -> None:
        """Solution not provided in SIF file"""
        return None

    @property
    def expected_objective_value(self) -> None:
        """Optimal value for N=1500 from SIF file"""
        # Not provided for N=1500, but pattern suggests around 2.87
        return None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class TRIGON2NE(AbstractNonlinearEquations):
    """TRIGON2NE problem - SCIPY global optimization benchmark example Trigonometric02.

    Fit: y = (0, sqrt(8)*sin(7*(x_i-0.9)^2), sqrt(6)*sin(14*(x_i-0.9)^2), x_i) + e

    Nonlinear-equation formulation of TRIGON2.SIF

    Source: Problem from the SCIPY benchmark set
    https://github.com/scipy/scipy/tree/master/benchmarks/benchmarks/go_benchmark_functions

    SIF input: Nick Gould, Jan 2020

    Classification: NOR2-MN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    _n: int = 10

    def __init__(self, n: int = 10):
        self._n = n

    @property
    def n(self):
        """Number of variables."""
        return self._n

    @property
    def m(self):
        """Number of equations: 3*n + 1."""
        return 3 * self._n + 1

    def constraint(self, y):
        """Compute the nonlinear equations."""
        x = y
        n = self.n

        sqrt6 = jnp.sqrt(6.0)
        sqrt8 = jnp.sqrt(8.0)

        # Vectorized computation
        d = x - 0.9
        y_val = d * d
        fb_vals = sqrt8 * jnp.sin(7.0 * y_val) + sqrt6 * jnp.sin(14.0 * y_val)
        fc_vals = jnp.zeros(n)  # FC(i) = 0
        fd_vals = d  # FD(i) = x_i - 0.9

        # Interleave FB, FC, FD values: FB(0), FC(0), FD(0), FB(1), FC(1), FD(1), ...
        interleaved = jnp.stack([fb_vals, fc_vals, fd_vals], axis=1).flatten()

        # Prepend FA = -1
        equations = jnp.concatenate([jnp.array([-1.0]), interleaved])

        return equations, None  # No inequality constraints

    @property
    def bounds(self):
        """No explicit bounds."""
        return None

    @property
    def y0(self):
        """Initial guess: x_i = i/n for i = 1, ..., n."""
        n = self.n
        indices = jnp.arange(1, n + 1, dtype=jnp.float64)
        return indices / n

    @property
    def args(self):
        """No additional arguments required."""
        return None

    @property
    def expected_result(self):
        """Expected result - solution with all variables at optimal values."""
        # The optimal solution would have x_i = 0.9 to make FD(i) = 0
        return jnp.full(self.n, 0.9, dtype=jnp.float64)

    @property
    def expected_objective_value(self):
        """Expected objective value at the optimal solution."""
        # For nonlinear equations, this is typically 0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class ROSZMAN1LS(AbstractUnconstrainedMinimisation):
    """
    NIST Data fitting problem ROSZMAN1.

    Fit: y = b1 - b2*x - arctan[b3/(x-b4)]/pi + e

    Source:
    Problem from the NIST nonlinear regression test set
    http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    Reference:
    Roszman, L., NIST (1979).
    Quantum Defects for Sulfur I Atom.

    SIF input: Nick Gould and Tyrone Rees, Oct 2015

    classification SUR2-MN-4-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0, 1})

    n: int = 4  # 4 parameters
    m: int = 25  # 25 data points

    @property
    def y0(self):
        if self.y0_iD == 0:
            # START1
            return jnp.array([0.1, -0.00001, 1000.0, -100.0])
        else:
            # START2
            return jnp.array([0.2, -0.000005, 1200.0, -150.0])

    @property
    def args(self):
        # X data
        x_data = jnp.array(
            [
                -4868.68,
                -4868.09,
                -4867.41,
                -3375.19,
                -3373.14,
                -3372.03,
                -2473.74,
                -2472.35,
                -2469.45,
                -1894.65,
                -1893.40,
                -1497.24,
                -1495.85,
                -1493.41,
                -1208.68,
                -1206.18,
                -1206.04,
                -997.92,
                -996.61,
                -996.31,
                -834.94,
                -834.66,
                -710.03,
                -530.16,
                -464.17,
            ]
        )

        # Y data
        y_data = jnp.array(
            [
                0.252429,
                0.252141,
                0.251809,
                0.297989,
                0.296257,
                0.295319,
                0.339603,
                0.337731,
                0.333820,
                0.389510,
                0.386998,
                0.438864,
                0.434887,
                0.427893,
                0.471568,
                0.461699,
                0.461144,
                0.513532,
                0.506641,
                0.505062,
                0.535648,
                0.533726,
                0.568064,
                0.612886,
                0.624169,
            ]
        )

        return (x_data, y_data)

    def objective(self, y, args):
        b1, b2, b3, b4 = y
        x_data, y_data = args

        # Model: y = b1 - b2*x - arctan[b3/(x-b4)]/pi
        # Note: The SIF file element E7 computes -ATAN(V1/(V2-X))/PI
        # where V1=B3, V2=B4, X=x_data
        # This means the term is -arctan(b3/(b4-x))/pi, not -arctan(b3/(x-b4))/pi
        # However, since arctan is an odd function: arctan(-z) = -arctan(z)
        # So -arctan(b3/(b4-x))/pi = -arctan(-b3/(x-b4))/pi = arctan(b3/(x-b4))/pi
        model_vals = b1 - b2 * x_data + jnp.arctan(b3 / (x_data - b4)) / jnp.pi
        residuals = model_vals - y_data

        # Sum of squared residuals
        return jnp.sum(residuals**2)

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # The optimal objective value is not explicitly given in the SIF file
        return None



import jax
import jax.numpy as jnp

from ..._problem import AbstractConstrainedQuadraticProblem


class NCVXQP4(AbstractConstrainedQuadraticProblem):
    """NCVXQP4 problem - a non-convex quadratic program.

    A non-convex quadratic program.

    SIF input: Nick Gould, April 1995

    Classification: QLR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 10000  # Default size

    @property
    def m(self):
        """Number of constraints."""
        return self.n // 4  # M = N/4

    def objective(self, y, args):
        """Compute the objective."""
        del args

        n = self.n
        x = y

        # The objective is a sum of quadratic terms
        # For each i from 1 to n:
        # OBJ(i) = 0.5 * p * (x[i] + x[mod(2i-1,n)+1] + x[mod(3i-1,n)+1])^2
        # where p = i for i <= n/4 and p = -i for i > n/4 (25% positive eigenvalues)

        def compute_term(i):
            # Positions (0-indexed)
            i1 = i
            i2 = (2 * (i + 1) - 1) % n
            i3 = (3 * (i + 1) - 1) % n

            # Sum the variables
            sum_vars = x[i1] + x[i2] + x[i3]

            # Parameter p (25% positive eigenvalues)
            nplus = n // 4
            p = jnp.where(i < nplus, jnp.float64(i + 1), jnp.float64(-(i + 1)))

            # Compute 0.5 * p * sum^2
            return 0.5 * p * sum_vars * sum_vars

        indices = jnp.arange(n)
        terms = jax.vmap(compute_term)(indices)
        return jnp.sum(terms)

    def constraint(self, y):
        """Compute the constraints.

        Each constraint i has the form:
        x[i] + 2*x[mod(4i-1,n)+1] + 3*x[mod(5i-1,n)+1] = 6
        """
        n = self.n
        m = self.m
        x = y

        def compute_constraint(i):
            # Positions (0-indexed)
            i1 = i
            i2 = (4 * (i + 1) - 1) % n
            i3 = (5 * (i + 1) - 1) % n

            # Compute x[i] + 2*x[j] + 3*x[k] - 6
            return x[i1] + 2.0 * x[i2] + 3.0 * x[i3] - 6.0

        indices = jnp.arange(m)
        equalities = jax.vmap(compute_constraint)(indices)

        return equalities, None

    @property
    def y0(self):
        """Initial guess."""
        return jnp.full(self.n, 0.5)

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return ()

    @property
    def bounds(self):
        """Variable bounds: 0.1 <= x[i] <= 10.0 for all i."""
        n = self.n
        lower = jnp.full(n, 0.1)
        upper = jnp.full(n, 10.0)
        return lower, upper

    @property
    def expected_result(self):
        """Expected result based on problem name."""
        raise NotImplementedError("Expected result not available for NCVXQP4")

    @property
    def expected_objective_value(self):
        """Expected objective value at the solution."""
        return None  # Not available



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class ARGLINA(AbstractUnconstrainedMinimisation):
    """ARGLINA function.

    Variable dimension full rank linear problem.

    Source: Problem 32 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#80 (with different N and M)
    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 200  # SIF file suggests 10, 50, 100, or 200
    m: int = 400  # SIF file suggests m >= n and values like 20, 100, 200, or 400

    def objective(self, y, args):
        del args
        n = self.n
        m = self.m

        # Based on the AMPL model in arglina.mod
        # First n residuals: for i in 1..N
        # (sum{j in 1..i-1} -2*x[j]/M) + x[i]*(1-2/M) +
        # (sum {j in i+1..N} -2*x[j]/M) - 1

        # Sum of all x values scaled by -2/m
        total_sum = jnp.sum(y) * (-2.0 / m)

        # First n residuals
        # Each residual i = total_sum + y[i] * (1 - (-2/m)) - 1
        # = total_sum + y[i] * (1 + 2/m) - 1
        # = total_sum + y[i] - 1
        first_n_residuals = total_sum + y - 1.0

        # Remaining m-n residuals: all equal to total_sum - 1
        remaining_residuals = jnp.full(m - n, total_sum - 1.0)

        # Combine all residuals
        residuals = jnp.concatenate([first_n_residuals, remaining_residuals])

        # Sum of squares of residuals
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        # Initial value of 1.0 as specified in the SIF file
        return jnp.ones(self.n)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return None



import jax
import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractConstrainedQuadraticProblem


class CVXQP1(AbstractConstrainedQuadraticProblem):
    """CVXQP1 problem - a convex quadratic program.

    A convex quadratic program with sparse structure.

    SIF input: Nick Gould, May 1995

    Classification: QLR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 100  # Default size

    @property
    def m(self):
        """Number of constraints."""
        return self.n // 2  # M = N/2

    def objective(self, y, args):
        """Compute the objective."""
        del args

        n = self.n
        x = y

        # The objective is a sum of quadratic terms
        # For each i from 1 to n:
        # OBJ(i) = 0.5 * i * (x[i] + x[mod(2i-1,n)+1] + x[mod(3i-1,n)+1])^2

        def compute_term(i):
            # Positions (0-indexed)
            i1 = i

            # For mod(2i-1, n) + 1 in SIF notation:
            # i is 0-indexed here, so (i+1) is 1-indexed
            i2 = (2 * (i + 1) - 1) % n  # This gives 0-indexed position

            # For mod(3i-1, n) + 1 in SIF notation:
            i3 = (3 * (i + 1) - 1) % n  # This gives 0-indexed position

            alpha = x[i1] + x[i2] + x[i3]
            p = inexact_asarray(i + 1)  # P parameter is i (1-indexed)

            return 0.5 * p * alpha * alpha

        # Use vmap to vectorize over all indices
        terms = jax.vmap(compute_term)(jnp.arange(n))
        obj = jnp.sum(terms)

        return obj

    def constraint(self, y):
        """Compute the constraints."""
        n = self.n
        m = self.m
        x = y

        # Linear equality constraints
        # For each i from 1 to m:
        # CON(i) = x[i] + 2*x[mod(4i-1,n)+1] + 3*x[mod(5i-1,n)+1] = 6

        def compute_constraint(i):
            # Positions (0-indexed)
            i1 = i

            # For mod(4i-1, n) + 1 in SIF notation:
            i2 = (4 * (i + 1) - 1) % n  # This gives 0-indexed position

            # For mod(5i-1, n) + 1 in SIF notation:
            i3 = (5 * (i + 1) - 1) % n  # This gives 0-indexed position

            c = x[i1] + 2.0 * x[i2] + 3.0 * x[i3] - 6.0
            return c

        # Use vmap to vectorize over all constraint indices
        constraints = jax.vmap(compute_constraint)(jnp.arange(m))

        return constraints, None

    def equality_constraints(self):
        """All constraints are equalities."""
        return jnp.ones(self.m, dtype=bool)

    @property
    def y0(self):
        """Initial guess."""
        # Default value is 0.5
        return inexact_asarray(jnp.full(self.n, 0.5))

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def bounds(self):
        """Variable bounds."""
        # 0.1 <= x[i] <= 10.0 for all i
        lower = jnp.full(self.n, 0.1)
        upper = jnp.full(self.n, 10.0)
        return lower, upper

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # From comment in SIF file for n=1000
        # We don't have the value for n=100
        return None



from jax import numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractNonlinearEquations


# TODO: Human review needed
# Attempts made:
# 1. Vectorized the for-loop implementation
# 2. Ensured dtype consistency throughout
# 3. Used proper dtype casting for all constants
# Suspected issues: Numerical precision in Chebyshev polynomial calculation
# Resources needed: Detailed analysis of Fortran implementation
class PALMER5ENE(AbstractNonlinearEquations):
    """A nonlinear least squares problem
    arising from chemical kinetics.

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 T_0 + A2 T_2 + A4 T_4 + A6 T_6 + A8 T_8 +
                 A10 T_10 + L * EXP( -K X**2 )
    where T_i is the i-th (shifted) Chebyshev polynomial

    Source:
    M.  Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1992.
    Bound-constrained nonlinear equations version: Nick Gould, June 2019.

    classification NOR2-RN-8-12
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def residual(self, y, args) -> Float[Array, "12"]:
        """Residual function for the nonlinear equations."""
        a0, a2, a4, a6, a8, a10, k, l = y

        # X data (radians) - only indices 12-23 from original problem
        x_data = jnp.array(
            [
                0.000000,
                1.570796,
                1.396263,
                1.308997,
                1.221730,
                1.125835,
                1.047198,
                0.872665,
                0.698132,
                0.523599,
                0.349066,
                0.174533,
            ],
            dtype=y.dtype,
        )

        # Y data (KJmol-1)
        y_data = jnp.array(
            [
                83.57418,
                81.007654,
                18.983286,
                8.051067,
                2.044762,
                0.000000,
                1.170451,
                10.479881,
                25.785001,
                44.126844,
                62.822177,
                77.719674,
            ],
            dtype=y.dtype,
        )

        # Compute shifted Chebyshev polynomials vectorized
        # T_0 = 1
        # Y = 2*(X - A)/(B - A) - 1 where A = -pi/2, B = pi/2
        # So Y = 2*X/pi
        pi = jnp.pi
        a_val = -pi / 2
        b_val = pi / 2
        diff = b_val - a_val  # = pi

        # Vectorized computation of y_shift for all data points
        y_shift = jnp.array(2.0, dtype=y.dtype) * (x_data - a_val) / diff - jnp.array(
            1.0, dtype=y.dtype
        )  # shape (12,)

        # Initialize Chebyshev polynomials for all data points
        # t[j] will have shape (12,) for j-th polynomial evaluated at all points
        t = []
        t.append(jnp.ones_like(x_data))  # T_0 = 1
        t.append(y_shift)  # T_1 = y_shift

        # Compute T_2 through T_10 via recursion
        for j in range(2, 11):  # j = 2, 3, ..., 10
            t_j = jnp.array(2.0, dtype=y.dtype) * y_shift * t[j - 1] - t[j - 2]
            t.append(t_j)

        # Model: Y = A0*T_0 + A2*T_2 + A4*T_4 + A6*T_6 + A8*T_8 + A10*T_10
        #          + L*exp(-K*X^2)
        model = (
            a0 * t[0]
            + a2 * t[2]
            + a4 * t[4]
            + a6 * t[6]
            + a8 * t[8]
            + a10 * t[10]
            + l * jnp.exp(-k * x_data * x_data)
        )

        residuals = model - y_data

        return residuals

    @property
    def y0(self) -> Float[Array, "8"]:
        """Initial guess for the optimization problem."""
        # From START POINT in SIF
        return jnp.array(
            [
                1.9264e01,  # A0
                -1.7302e00,  # A2
                4.0794e01,  # A4
                8.3021e-01,  # A6
                3.7090e00,  # A8
                -1.7723e-01,  # A10
                10.0,  # K
                1.0,  # L (default)
            ]
        )

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file doesn't provide a solution
        return None

    @property
    def expected_objective_value(self) -> Float[Array, ""]:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    def num_constraints(self) -> tuple[int, int, int]:
        """Returns the number of constraints."""
        # 12 equality constraints (the residuals) + 1 finite bound (lower bound on K)
        num_equalities = 12
        num_inequalities = 0
        num_bounds = 1  # 1 finite lower bound on K
        return num_equalities, num_inequalities, num_bounds

    @property
    def bounds(self) -> tuple[Float[Array, "8"], Float[Array, "8"]]:
        """Bounds on variables."""
        # Lower bounds: K has implicit lower bound (likely 0)
        lower = jnp.array(
            [-jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, 0.0, -jnp.inf]
        )
        # Upper bounds: all inf
        upper = jnp.array(
            [jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf]
        )
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class PALMER2B(AbstractBoundedMinimisation):
    """A nonlinear least squares problem with bounds arising from chemical kinetics.

    model: H-N=C=O TZVP + MP2
    fitting Y to A2 X**2 + A4 X**4
                + B / ( C + X**2 ), B, C nonnegative.

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1990.

    classification SBR2-RN-4-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Data points
    M: int = 23

    # X values (radians)
    X_data = jnp.array(
        [
            -1.745329,
            -1.570796,
            -1.396263,
            -1.221730,
            -1.047198,
            -0.937187,
            -0.872665,
            -0.698132,
            -0.523599,
            -0.349066,
            -0.174533,
            0.0,
            0.174533,
            0.349066,
            0.523599,
            0.698132,
            0.872665,
            0.937187,
            1.047198,
            1.221730,
            1.396263,
            1.570796,
            1.745329,
        ]
    )

    # Y values (KJmol-1)
    Y_data = jnp.array(
        [
            72.676767,
            40.149455,
            18.8548,
            6.4762,
            0.8596,
            0.00000,
            0.2730,
            3.2043,
            8.1080,
            13.4291,
            17.7149,
            19.4529,
            17.7149,
            13.4291,
            8.1080,
            3.2053,
            0.2730,
            0.00000,
            0.8596,
            6.4762,
            18.8548,
            40.149455,
            72.676767,
        ]
    )

    @property
    def n(self):
        """Number of variables."""
        return 4  # A2, A4, B, C

    def objective(self, y, args):
        """Compute the objective function (least squares)."""
        del args

        # Extract variables
        A2, A4, B, C = y

        # Precompute powers of X
        X_sqr = self.X_data * self.X_data
        X_4 = X_sqr * X_sqr

        # Model predictions
        predictions = A2 * X_sqr + A4 * X_4 + B / (C + X_sqr)

        # Residuals
        residuals = predictions - self.Y_data

        # Sum of squares
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        """Initial guess."""
        return jnp.ones(4)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return jnp.zeros(4)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(1.7552259e-02)

    @property
    def bounds(self):
        """Returns the bounds on the variable y."""
        # A2, A4 are free, B and C have lower bounds
        lower = jnp.array([-jnp.inf, -jnp.inf, 0.00001, 0.00001])
        upper = jnp.full(4, jnp.inf)
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class DANIWOOD(AbstractNonlinearEquations):
    """NIST data fitting problem DANIWOOD as a nonlinear equations problem.

    This is a revised version of the original inaccurate
    formulation of DANWOOD, with corrections provided by Abel Siqueira,
    Federal University of Parana. The original DANWOOD problem in the SIF
    collection contains an incorrect formulation and should not be used

    Fit: y = b1*x**b2

    Source: Problem from the NIST nonlinear regression test set
    http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    Daniel, C. and F. S. Wood (1980).
    Fitting Equations to Data, Second Edition.
    New York, NY: John Wiley and Sons, pp. 428-431.

    SIF input: Nick Gould and Tyrone Rees, Oct 2015 (as DANWOOD)
              correction by Abel Siqueira, Feb 2019 (renamed DANIWOOD)

    Classification: NOR2-MN-2-6
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Data values
    x_data = jnp.array([1.309, 1.471, 1.490, 1.565, 1.611, 1.680])
    y_data = jnp.array([2.138, 3.421, 3.597, 4.340, 4.882, 5.660])

    @property
    def n(self):
        """Number of variables."""
        return 2

    def num_residuals(self):
        """Number of residuals."""
        return 6  # One residual per data point

    def residual(self, y, args):
        """The residuals: b1*x**b2 - y_i = 0"""
        del args
        b1, b2 = y

        # Model predictions: b1 * x**b2
        predictions = b1 * self.x_data**b2

        # Residuals
        residuals = predictions - self.y_data

        return residuals

    @property
    def y0(self):
        """Initial guess."""
        # Using START1 from SIF file
        return jnp.array([1.0, 5.0])

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution.

        From NIST reference:
        b1 = 0.76886226
        b2 = 3.8604055
        """
        return jnp.array([0.76886226, 3.8604055])

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (0 for constrained formulation)."""
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



from ._curly_base import CURLYBase


# TODO: this should still be compared against another CUTEst interface
class CURLY30(CURLYBase):
    """The CURLY 30 function.

    A banded function with semi-bandwidth 30 and
    negative curvature near the starting point.

    Source: Nick Gould.
    SIF input: Nick Gould, September 1997.

    Classification: OUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, n: int = 10000, k: int = 30):
        super().__init__(n=n, k=k)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class TOINTGSS(AbstractUnconstrainedMinimisation):
    """TOINTGSS problem - Toint's Gaussian problem.

    This problem has N-2 trivial groups, all of which have 1 nonlinear element.

    Source: problem 21 in
    Ph.L. Toint,
    "Test problems for partially separable optimization and results
    for the routine PSPMIN",
    Report 83/4, Department of Mathematics, FUNDP (Namur, B), 1983.

    SIF input: Ph. Toint, Dec 1989, corrected Nick Gould, July 1993.

    Classification: OUR2-AY-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    _n: int = 5000

    @property
    def n(self):
        """Number of variables."""
        return self._n

    def objective(self, y, args):
        """Compute the objective function."""
        del args
        x = y
        n = self.n

        # Parameter: 10 / (n-2)
        ap = 10.0 / (n - 2)

        # Extract all overlapping triplets using slicing
        # v1 = x[0:n-2], v2 = x[1:n-1], v3 = x[2:n]
        v1 = x[:-2]
        v2 = x[1:-1]
        v3 = x[2:]

        # Compute all elements vectorized
        u1 = v1 - v2
        u2 = v3

        alpha = 0.1
        u1sq = u1 * u1
        u2sq = u2 * u2
        t = alpha + u2sq
        au2sq = ap + u2sq

        expa = jnp.exp(-u1sq / t)
        tmexpa = 2.0 - expa

        # Element values
        elements = au2sq * tmexpa

        # Sum all elements
        return jnp.sum(elements)

    @property
    def y0(self):
        """Initial guess."""
        return jnp.full(self.n, 3.0)

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None  # Not provided in SIF

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return None  # Not provided in SIF
